#' Divide a set of objects into two homogeneous sets
#'
#' This function divides a given set of objects into two subsets
#' so that the weighted sum of the impurity measures of the subsets
#' are minimized using a greedy algorithm with Depth-First Search (DFS). \cr
#' Japanese help file: \link[lazy.tree]{Divide_JPH}
#'
#' @param ys Criterion variable vector
#' @param Xs Regressor matrix
#' @param Ys The dichotomized criterion matrix for \code{method=4}
#' @param S Index vector indicating the original location of
#' the rows of ys and X. \cr
#' (ys, Xs, Ys, S) have the same # of rows or of the same length.
#' @param impfunc impurity function
#' @param method Treatment of the categorical regressor variable. \cr
#' = 0 to dummy expand the categorical regressor, \cr
#' = 1 to use the first canonical varialbe, \cr
#' = 2 to use up to the 2nd canonical variables and their sum and difference,\cr
#' = 3 to use one-versus-rest dichotomization. \cr
#' = 4 to use dichotomization of the criterion variable on the basis of
#' all possible bipartitions of the levels of the criterion variable.
#' @param ncatx Vector consisting of the # of cagegories of the regressors
#' @param app = 0 not to use app (all possible partitions) \cr
#' = 1 to use app when ncaty >= 3 and ncatx >= 3 \cr
#' = 9 to use app awlays for categorical predictors.
#' @param maxapp Maximum # of categories of the regressor variables
#' for which app is used.
#' @param maxncut Maximum # of cut point (threshold) candidate
#' @param levelsy # of levels of the criterion
#' @param levelsx a list of the # of levels of the regressors
#'
#' @details
#'  \code{ncat=0} or \code{ncat=1} means the variable is continuous, \cr
#'  \code{ncat=2} means the variable is binary, \cr
#'  \code{ncat >= 3} means the variable is polytomous.
#'
#' @examples
#' Xc <- data.frame(x1=as.factor(square$x1),x2=as.factor(square$x2))
#' Divide( square$Class4, Xc, impfunc=gini, S=1:nrow(Xc), method=1
#'       , ncatx=c(7,5), app=0 )
#' Divide( square$Class4, Xc, impfunc=gini, S=1:nrow(Xc), app=9
#'       , ncatx=c(7,5) )
#'
#'
#' @return
#' A list consisting of the following:
#' \preformatted{
#' j:         the variable number used to divide the set
#' threshold: the threshold of variable j when it is continuous, or NA.
#' valxj: the set of values of nominal variable j to be classified as left
#' valxjR: the set of values of nominal variable j to be classified as right
#'        Note that these are numeric variables, not factors.
#' k:     quantified variable number used or 0
#' nqx:   # of unique quantified variables
#' locLL and locRR: the location of obs in 1:nrow(Xs) to be classficed
#' left:  the set of observations classified as left
#' right: the set of observations classified as right
#' left.score:  The nobs weighted impurity measure of left set
#' right.score: The nobs weighted impurity measure of right set
#' score:       The sum of the above two.
#' }
#'
#'
#'
#' @export
#'

Divide <- function( ys, Xs, Ys=NULL, impfunc, S, method=1, ncatx=0
                  , app=0, maxapp=10, maxncut=50
                  , levelsy=NULL, levelsX=NULL ){
  # Modified version of "branch" function in Joe Suzuki(2020)
  # Shin-ichi Mayekawa
  # 20230901cot
  # use unique values, return threshold: 20230902cot,03,04
  # use quantify_x: 20230905
  # return th for categorical x: 20230905,06
  # X,y -> y,X: 20230909
  # make categorical xsj numeric: 20230910
  # branch renamed as divide: 20230910
  # method: 20230911,14
  # x < th -->  x <= th: 20230914
  # convert all factors to numeric: 20230914
  # always define uxsj0: 20230915
  # unweighted impurity function: 20230916
  # ncatx[j] >= 3 && lenuxsj >= 3: 20230919,20
  # instead of (y,X), pass (ys, Xs): 20230920
  # app (all possible partitions): 20230926dnc,27
  # maxapp: 20230927
  # candisc flag: 20230928
  # app=9: 20230928,30,1001
  # kk: 20230929
  # score dividec by N: 20230930
  # Y matrix: 20231005
  # maxncut: 20231006,14
  # rank conversion of uxsj: 20231014
  # score NOT divided by N: 20231014
  # output nqx: 20231109
  # rank ties.method="min" and roxygen2: 20231120
  # levels: 20231122
  # factor removed: 20231123
  # uxsjk: sort, remove max, unique quantile: 20231219dnc
  # nobsL and nobsR clarified: 20231231
  # set threshold=NA when ncatx=0: 20240210
  # valxjR: 20240211
  # reflect uxsj: 20240229,0301
  # do NOT set threshold=NA when ncatx=0: 20240229
  # return locLL and locRR: 20240329dnc
  #

  # Args:
  #  ys       y vector of those observations in region S
  #  Xs       X matrix of those observations in region S
  #  impfunc  impurity function
  #  S        a vector of observation ids in region S
  #  method   method for quantifying x variable
  #  ncatx    a vector of the number of categories of X variables:
  #  0, 2, 3, ...
  #
  #

  #
  # Original version named as branch is found in: Joe Suzuki (2020)
  #
  # Major modifications:
  #
  # 1) instead of all obs in S, the ones with unique x values will be
  # used as the threshold.
  #
  # 2) threshold will be returned in info instead of i.
  # Note that value of i returned to DecTree is used only to
  # find the threshold.
  #
  # 3) vectorized comparison
  #
  # 4) use quantify_x to quantify categorical regressors
  #
  #

  # variable type by # of categories
  # ncat=0 or 1  continuous
  # ncat=2 binary
  # ncat > 2 polytomous

  # maxapp
  if( maxapp > 54 ) maxapp=54

  # # of obs in S and # of x variables
  n=length(S)
  nxvar=ncol(Xs)
  if( n == 0 ) return(NULL)

  if( is.factor(ys) ) ncaty=length(unique(ys))
  else ncaty=0

  best.score=Inf


  # each variable
  for( j in 1:nxvar ){


    # all of variable j in the region of interest indexed by S.
    # xsj=X[S,j]
    xsj=Xs[,j]

    # make categorical xsj numeric so that quantify_x can handle it.
    if( is.factor(xsj) ) xsj=as.numeric(xsj)     # factor removed: 20231123

    # unique values of xsj.
    lenxsj=length(xsj)
    temp=Unique( xsj )
    uxsj=matrix(temp$value,,1)
    uxsj0=uxsj
    lenuxsj=nrow(uxsj)
    uloc=temp$loc       # location of uxsj in xsj


    # Here,
    #  xsj     is the j-th column of Xs,
    #  uxsj    is the vector of unique values of xsj,
    #  lenuxsj is the # of unique values of xsj,
    #  uloc    is the list of locations of uxsj in xsj.
    #
    #  Let ux is a unique value of interest. locLL
    #   locL=which(uxsj==ux)                 # locations, in uxsj, of ux
    #   locLL=sort( unlist(uloc[locL]) )     # locations, in S, of ux
    #   left=S[locLL]                        # id of obs which are equal to ux.
    #
    #  Therefor, S[uloc[i]] gives X[S[uloc[[i]]],j] == uxsj[i]
    #


    candisc=0
    nobsL=0
    nobsR=0


    if( ( app==9 && lenuxsj <= maxapp && ncatx[j] >= 2 )
        || ( app==1 && lenuxsj <= maxapp && ncatx[j] >= 3
             && length(unique(ys)) >= 3 )                  ){

      # All Possible Partitions

      nitems=lenuxsj

      # partition
      # Schatzoff, M., Tsao, R.  and Fienberg, S. (1968)
      #
      n21=2^(nitems-1)-1+.00001
      x=rep(0,nitems) # location of 1s
      mode(x)="logical"
      for( k in 1:n21){
        i=1; kk=k;
        while( kk %% 2 == 0 ){ i=i+1; kk=kk/2 }
        x[i]=!x[i]

        # Here, x[i]=1, i=1,2, ...,lenuxsj,  means that, in the k-th partition,
        # the i-th level of the j-th regressor is to be classified as left.

        #   Print( k, kk,i, paste(x,collapse=",") )

        locL=which( x )
        locR=which( !x )
        left=NULL;  right=NULL
        if(length(locL) > 0 ){
          locLL=sort( unlist(uloc[locL]) )
          left=S[locLL]
          nobsL=length(left)
        }
        else locLL=NULL
        locRR=setdiff( 1:lenxsj, locLL )
        if( length(locRR) > 0 ){
          right=S[locRR]
          nobsR=length(right)
        }
        else locRR=NULL

        # impurity measures
        L=nobsL*impfunc(ys[locLL])
        R=nobsR*impfunc(ys[locRR])
        score=L+R

        if( score < best.score ){

          best.score <- score

          # find the threshold category values of categorical x
          if( ncatx[j] <= 1 ){
            valxj=NULL; valxjR=NULL
          }
          else{
            valxj=uxsj0[locL]; valxjR=uxsj0[locR]
            valxj=sort(valxj); valxjR=sort(valxjR)
          }

          info <- list(threshold=NA, j=j, k=NA, nqx=0
                       , left=left, right=right, locLL=locLL, locRR=locRR
                       , score=best.score, left.score=L, right.score=R
                       , valxj=valxj, valxjR=valxjR )

        } # score check

      } # end of k loop


    } # end of app=1
    else{


      # use quantification of categorical predictors, if any.
      #
      # Conditions for quantification:
      #  app=0 or lenuxsi > 54
      #  or ys is binary or continuous   or  xj is binary or continuous,
      #

      if( ncaty > 0   &&  ncatx[j] >= 3 && lenuxsj >= 3 ){
        # quantify the unique values of xsj when y is categorical
        temp2=quantify_x( ys, xsj, uxsj, uloc, Ys, method=method )
        uxsj=temp2$newux
        lenuxsj=nrow(uxsj)
        candisc=temp2$candisc

        # convert the result to rank and drop redundant columns
        uxsj=apply( uxsj, 2, rank, ties.method="min" )
        uxsj=t(unique(t(uxsj)))

        # reflect uxsj                   20240229
#        cc=cor(1:lenuxsj,uxsj)
#        locnegc=which( cc < 0 )
#        if( length(locnegc) > 0 )  uxsj[,locnegc]=-uxsj[,locnegc]

        if( ncol(uxsj) >= 2 ){
          R=cor(cbind(1:nrow(uxsj),uxsj))
          # Print(R)
          ev=eigen(R)
          eve=ev$vectors[,1]
          eve1=eve[1]
          eve=sign(eve1)*eve[-1]
          locn=which(eve < 0)
          if( length(locn) > 0 ){
            uxsj[,locn]=-uxsj[,locn]
          }
        }

if(0){
cat("\n")
Print(j)
Print(uxsj)
Print(cc)
}

      }

      nqx=ncol(uxsj)

      # quantify_x may return  uxsj with 0 columns!  20231219
      if( nqx == 0 ) next  # j loop

      # for each unique value , uxsj_i,  of xsj as candidate threshold
      # compare it to the unique values, uxsj,  of xsj.
      # Note that the comparison is between unique values.
      # The location in xsj of the unique values is recorded in loc list.
# if( ncol(Xs) == 1 ) Print(nqx,uxsj)

      # loop for quantified categorical regressor
      for( k in 1:nqx ){

        # maxncut 20231006, # sort, unique: 20231219dnc
        uxsjk=sort( unique( uxsj[,k] ) )
        if( length(uxsjk) > 1 ) uxsjk=uxsjk[-length(uxsjk)]  # 20231219dnc
        if( length(uxsjk) > maxncut ){
          uxsjk=unique(quantile(uxsjk,probs=seq(0,1,length=maxncut+1),type=1))
          uxsjk=uxsjk[-maxncut]
        }

# if( ncol(Xs) == 1 ) Print(k, uxsj, uxsj[,k], uxsjk)
        # loop for the threshold candidate
        for(i in 1:length(uxsjk)){

          # uxsjk_i is the i-th threshold candidate.
          uxsjk_i=uxsjk[i]
# if( ncol(Xs) == 1 ) Print(k,i,uxsjk, uxsjk_i)
          # Divide uxsj at threshold and store the results in left and right.
          # Note that locL is the location in uxsj,
          # the unique values, not xsj.
          # The location in xsj is stored in temp$loc: that is,
          # xsj[ temp$loc[[i]] ] have the same value uxsj[i].
          # Finally, the location in xsj must be converted to
          # the location in xj via S.
          #
          locL=which( uxsj[,k] <= uxsjk_i )
          locR=which( uxsj[,k] > uxsjk_i )
# if( ncol(Xs) == 1 ) Print(locL,locR)
          left=NULL;  right=NULL
          if(length(locL) > 0 ){
            locLL=sort( unlist(uloc[locL]) )
            left=S[locLL]
            nobsL=length(left)
          }
          else locLL=NULL
          locRR=setdiff( 1:lenxsj, locLL )
          if( length(locRR) > 0 ){
            right=S[locRR]
            nobsR=length(right)
          }
          else locRR=NULL
# Print(nobsL,nobsR)
          # impurity measures
          L=nobsL*impfunc(ys[locLL])
          R=nobsR*impfunc(ys[locRR])
          score=L+R
# Print(j,k,i,score,best.score, score < best.score)
          if( score < best.score ){

            best.score <- score

            # find the threshold category values of categorical x
            if( ncatx[j] <= 1 ){
              valxj=NULL; valxjR=NULL
            }
            else{
              valxj=uxsj0[locL]; valxjR=uxsj0[locR]
              valxj=sort(valxj); valxjR=sort(valxjR)
            }
# Print(best.score,j,k,nqx,L,R)
#            if( ncatx[j] > 0 ) uxsjk_i=NA         # 20240210 -> 20240229

            kk=k
            info <- list(threshold=uxsjk_i, j=j, k=kk, nqx=nqx
                         , left=left, right=right, locLL=locLL, locRR=locRR
                         , score=best.score, left.score=L, right.score=R
                         , valxj=valxj, valxjR=valxjR )

          } # score check

        } # end of i for unique obs

      } # end of k loop


    } # end of app=0

#XLj=Xs[left,j]
#XRj=Xs[right,j]
#Print(j,uxsjk_i,k)
#Print(table(Xs[,j]))
#Print(table(XLj),table(XRj))


  } # end of j for variable

  return(info)

} # end of Divide



