#' Divide a set of objects into two homogeneous sets with resampling
#'
#' This function searches the average division of a given subset
#' by resmpling the subset.  \cr
#' Japanese help file: \link[lazy.tree]{DivideM_JPH}
#'
#' @param ys Criterion variable vector
#' @param Xs Regressor matrix
#' @param Ys The dichotomized criterion matrix for \code{method=4}
#' @param impfunc impurity function
#' @param S Index vector indicating the original location of
#' the rows of ys and X.
#' @param method Treatment of the categorical regressor variable. \cr
#' = 0 to dummy expand the categorical regressor, \cr
#' = 1 to use the first canonical varialbe, \cr
#' = 2 to use up to the 2nd canonical variables and their sum and
#' difference,\cr
#' = 3 to use one-versus-rest dichotomization. \cr
#' = 4 to use dichotomization of the criterion variable on the basis of
#' all possible bipartitions of the levels of the criterion variable.
#' @param ncatx Vector consisting of the # of cagegories of the regressors
#' @param app = 0 not to use app (all possible partitions) \cr
#' = 1 to use app when ncaty >= 3 and ncatx >= 3 \cr
#' = 9 to use app awlays for categorical predictors.
#' @param maxapp Maximum # of categories of the regressor variables
#' for which app is used.
#' @param maxncut Maximum # of cut point (threshold) candidate
#' @param levelsy # of levels of the criterion
#' @param levelsx a list of the # of levels of the regressors
#' @param majestic > 0 to use within node resampling: \cr
#' = 1 to use modal splitting variable \cr
#' = 2 to use binary LR to determine J and predict ys from J\cr
#' = 3 to use binary LR to determine J\cr
#' = 4 to use continuous LR\cr
#' = 5 to use logit of continuous LR
#' @param nrepM # of divisions to try for majestic tree
#' @param psampleM percentage of the sample for majestic tree
#' @param rf = 0 not to sample regressor variables
#' @param nxvarM # of variables to sample for majestic tree (Random Forest)
#' @param minLeftM minimum proportion of x values to be classified as Left
#' @param maxRightM maximum proportion of x values to be classified as Right
#' @param r node number
#' @param debug = 1 to print intermediate result
#'
#' @details
#'  \code{ncat=0} or \code{ncat=1} means the variable is continuous, \cr
#'  \code{ncat=2} means the variable is binary, \cr
#'  \code{ncat >= 3} means the variable is polytomous.
#'
#' @examples
#'
#' N=500
#' Xc=data.frame(x1=as.factor(square3$x1), x2=square3$x2)
#' yc=as.factor(square3$y)
#' Xc=Xc[1:N,]
#' yc=yc[1:N]
#' S=1:N
#' maxncatx=max( apply( Xc, 2, function(x){length(unique(x))} ) )
#'
#' ncaty=length(levels(yc))
#'
#'   id=gen01pat( ncaty, sort=-2, partition=1 ) == 1  # logical matrix
#'   npart=nrow(id)
#'   uy=unique(yc)
#'   Yc=matrix(0,N,npart)
#'   for(k in 1:npart ){
#'     Yc[,k]=yc %in% uy[id[k,]]
#'   }
#'
#'   nx=ncol(Xc)
#'   ncatx=rep(0,nx)
#'   for( j in 1:nx ){
#'     if( is.factor(Xc[,j]) ) ncatx[j]=length(levels(Xc[,j]))
#'   }
#'
#' res00=DivideM( yc, Xc, Y=Yc, impfunc=entropy, S, method=1, ncatx=ncatx
#'               , app=0, maxapp=10, maxncut=50
#'               , levelsy=NULL, levelsX=NULL
#'               , psampleM=100, nrepM=20, r=0 )
#'
#'
#' @return
#' A list consisting of the following:
#' \preformatted{
#' j:         the variable number used to divide the set
#' threshold: the threshold of variable j when it is continuous
#' valxj: the set of values of nominal variable j to be classified as left.
#'        Note that this is a numeric variable, not a factor.
#' k:     quantified variable number used or 0
#' nqx:   # of unique quantified variables
#' left:  the set of observations classified as left
#' right: the set of observations classified as right
#' left.score:  The nobs weighted impurity measure of left set
#' right.score: The nobs weighted impurity measure of right set
#' score:       The sum of the above two.
#' }
#'
#'
#'
#' @export
#'

DivideM <- function( ys, Xs, Ys=NULL, impfunc, S, method=1, ncatx=0
                  , app=0, maxapp=10, maxncut=50
                  , levelsy=NULL, levelsX=NULL
                  , majestic=3, nrepM=50, psampleM=100, rf=1
                  , nxvarM=if(!is.null(ys) && !is.factor(ys))
                    max(floor(ncol(Xs)/3), 1) else floor(sqrt(ncol(Xs)))
                  , minLeftM=0.5, maxRightM=0.5, r=0, debug=0 ){
  # Modified version of "branch" function in Joe Suzuki(2020)
  # Shin-ichi Mayekawa
  # 20230901cot
  # use unique values, return threshold: 20230902cot,03,04
  # use quantify_x: 20230905
  # return th for categorical x: 20230905,06
  # X,y -> y,X: 20230909
  # make categorical xsj numeric: 20230910
  # branch renamed as divide: 20230910
  # method: 20230911,14
  # x < th -->  x <= th: 20230914
  # convert all factors to numeric: 20230914
  # always define uxsj0: 20230915
  # unweighted impurity function: 20230916
  # ncatx[j] >= 3 && lenuxsj >= 3: 20230919,20
  # instead of (y,X), pass (ys, Xs): 20230920
  # app (all possible partitions): 20230926dnc,27
  # maxapp: 20230927
  # candisc flag: 20230928
  # app=9: 20230928,30,1001
  # kk: 20230929
  # score dividec by N: 20230930
  # Y matrix: 20231005
  # maxncut: 20231006,14
  # rank conversion of uxsj: 20231014
  # score NOT divided by N: 20231014
  # output nqx: 20231109
  # rank ties.method="min" and roxygen2: 20231120
  # levels: 20231122
  # factor removed: 20231123
  # uxsjk: sort, remove max, unique quantile: 20231219dnc
  # majestic: 20231223cot,25,27
  # .Random.seed: 20231228,30
  # as.numeric(X[,J]): 20231230
  # mean threshold fix: 20240111
  # negative seedM: 20240111
  # valxjR: 20240211
  # when locLL is empty: 20240227dnc
  # majestic as option and majestic=1 method: 20240228
  # new negative seedM: 20240305dnc
  # use get and assign to handle .Random.seed: 20240309
  # treatment of .Random.seed moved to DecTree: 20240309
  # no seedM version: 20240313
  # use locLL and locRR from Divide: 20240329dnc,30,31
  # majestic=2, plotM deleted: 20240329
  # only Jall is saved: 20240401
  # new majestic=2: 20240406cot
  # bugfix for binary X: 20240407cot
  # majestic=3: 20240407cot
  # majestic=4: 20240409
  # majestic=4 bug fix and majestic=5: 20240412
  # bug fix for score calculation when method>=3: 20240416
  # majestic=-1: 20240416,17
  # yhatall renamed as sumyhat, Reflect: 20240417
  # RF: 20240425,27
  # bugfix for rf ncatx[locv]: 20240430
  #

  # Args:
  #  ys       y vector of those observations in region S
  #  Xs       X matrix of those observations in region S
  #  impfunc  impurity function
  #  S        a vector of observation ids in region S
  #  method   method for quantifying x variable
  #  ncatx    a vector of the number of categories of X variables:
  #  0, 2, 3, ...
  #
  #

# Print("********* Top of DivideM ***************")

  # variable type by # of categories
  # ncat=0 or 1  continuous
  # ncat=2 binary
  # ncat > 2 polytomous

  # maxapp
  if( maxapp > 54 ) maxapp=54

  # # of obs in S and # of x variables
  nobs=length(S)
  nxvar=ncol(Xs)
  if( nobs == 0 ) return(NULL)

  if( is.factor(ys) ) ncaty=length(unique(ys))
  else ncaty=0


  # resampling of (ys, Xs, Ys, S)
  #

  if( psampleM > 100 ) psampleM=100
  else if( psampleM < 0 ) psampleM=0


  # set random number seed
#  if( seedM != 0 ){
#    set.seed(abs(seedM))
#    .R.s.M=.Random.seed
#  }



  #####################################
  ############# start resampling ######
  #####################################

  # storage
  Jall=numeric(nrepM)
  THall=numeric(nrepM)
  Valxjall=vector(mode="list", length=nrepM)
  ValxjRall=vector(mode="list", length=nrepM)
  Reflect=numeric(nrepM)

  # resampling
  for( rep in 1:nrepM ){

   # cols
   if( rf != 0  &&  nxvarM <= nxvar ){
    locv=sample(1:nxvar, nxvarM, replace=0)
   }
   else{
    locv=1:nxvar
   }

   # rows
   loc=sample(1:nobs, nobs*psampleM/100, replace=1)

#   yss=ys[loc]
#   Xss=Xs[loc,locv,drop=0]
#   Ss=S[loc]
#   Yss=Ys[loc,,drop=0]

   res=Divide( ys[loc], Xs[loc,locv,drop=0]
                  , Ys[loc,,drop=0], impfunc=impfunc
                  , S=S[loc], method=method, ncatx=ncatx[locv]
                  , app=app, maxapp=maxapp, maxncut=maxncut
                  , levelsy=levelsy, levelsX=levelsX[locv] )


if( debug > 0 ){
  Print(res$j,locv[res$j], res$th,length(res$valxj),length(res$valxjR))
}

   # res$j     the variable number used to divide the set
   # res$th    the threshold of variable j when it is continuous
   # res$valxj the set of values of nominal variable j to be classified
   # res$valxjR the set of values of nominal variable j to be classified
   # as right
   # res$k     quantified variable number used.
   # res$nqx   # of unique quantified variable
   # res$left  the set of observations classified as left
   # res$right the set of observations classified as right
   # rest$left.score  The nobs weighted impurity measure of left set
   # rest$right.score The nobs weighted impurity measure of right set
   # rest$score       The sum of above two.


   Jall[rep]=locv[ res$j ]
   THall[rep]=res$th
   Valxjall[rep]=list(res$valxj)
   ValxjRall[rep]=list(res$valxjR)

  } # end of rep loop


  ###########################################################
  ############# end of resampling ###########################
  ############# Here, Jall, THall, Valxjall are ready. ######
  ###########################################################


  if( debug > 0 ) Print("End of Sampling in DivideM.")
  if( debug > 0 ){
   Print("\nDivideM",r)
   Print(table(Jall))
   # Sys.sleep(10)
  }


  #
  # post processing of resampling result
  #

  if( majestic == 1 ){

   # majestic = 1

   # modal value of regressor number j
   J=Mode(Jall)
   locJ=which(Jall==J)
   nrepM1=length(locJ)

   if( 0  &&  debug > 0 ){
    Print(J,locJ, ncatx[J])
    Print(summary(Xs[,J]))
    save(ys,Xs,Ys,S,ncatx, levelsX, J
         , file=paste("RpgmORG/d_",r,".RData",sep=""))
    Print("Saved!!", r)
   }


   # pick up one tree:   always use majestic=1

   res=Divide( ys, Xs[,J,drop=0]
               , Ys, impfunc=impfunc
               , S=S, method=method, ncatx=ncatx[J]
               , app=app, maxapp=maxapp, maxncut=maxncut
               , levelsy=levelsy, levelsX=levelsX[J] )

   # Print(res[c("j","th","valxj")])

   TH=res$threshold
   k=res$k
   nqx=res$nqx
   left=res$left
   right=res$right
   L=res$left.score
   R=res$right.score
   score=res$score
   Valxj=res$valxj
   ValxjR=res$valxjR
   locLL=res$locLL
   locRR=res$locRR

   if( debug > 0 ){
    Print("In DivideM")
    XLj=Xs[left,J]
    XRj=Xs[right,J]
    Print(table(Xs[,J]))
    Print(J,TH,k)
    Print(table(XLj),table(XRj))
   }

  } # end of majestic=1
  else{

   # majestic = 2 or 3 or 4 or 5
   #
   # For each replication:
   #  Calculate yhat as L or R using Jall[rep] and THall[rep] or Valxjall[rep].
   #    yhat[i]=1 if the i-th obs at rep-th replication is classified as L.
   #  Add yhat to sumyhat so that sumyhat[i] is the frequency of i-th obs
   #  being classfied as L.
   #
   # At the end of rep loop, sumyhat=sumyhat/nrepM contains
   # the relative frequency of the i-th obs beinc classified as L.
   #
   # From sumyhat, we use the following method to find J and TH:
   #
   #  majestic=2
   #   Dichotomize sumyhat at 0.5 and use it as the criterion to DecTree
   #   to find J.
   #   Then, using the original ys as the criterion and Xs[,J] as the regressor,
   #   run DecTree again to find the threshold.
   #
   #  majestic=3
   #   Dichotomize sumyhat at 0.5 and use it as the criterion to DecTree
   #   to find J and the threshold.
   #   Then, calculate score in terms of ys.
   #
   #  majestic=4
   #   Use sumyhat as the continuous criterion to DecTree to find J.
   #   Then, calculate score in terms of ys.
   #
   #  majestic=5
   #   Use logit(sumyhat) as the continuous criterion to DecTree.
   #   Then, calculate score in terms of ys.
   #
   #  majestic=-1
   #  Use dichotomized sumyhat at 0.5 to classify each obs to L and R.
   #  Set J=-1 and TH=NA.
   #

   sumyhat=numeric(nobs)
   yhat=numeric(nobs)
# YHA=NULL

   for( rep in 1:nrepM ){

    yhat=numeric(nobs)
    j=Jall[rep]
    if( ncatx[j] <= 2 ){
      yhat[ as.numeric(Xs[,j]) <= THall[rep] ]=1
    }
    else{
     Vj=Valxjall[[rep]]
     for( a in 1:length(Vj) ){
      yhat[ as.numeric(Xs[,j]) == Vj[a] ]=1
     }
    }
# YHA=cbind(YHA,yhat)
    if( rep == 1 ) sumyhat=yhat
    else{
     c=cor(sumyhat,yhat)
     if( !is.na(c)  &&  c < 0 ){
       yhat=1-yhat
       Reflect[rep]=1
     }
     sumyhat=sumyhat+yhat
    }

   } # end of rep loop

   # aggregated yhat:  Left or Right
   sumyhat=sumyhat/nrepM
   sumyhat5=as.factor(as.numeric(sumyhat >= 0.5))

   if( debug > 0 ){
     colnames(YHA)=Jall
     rsYHA=rowSums(YHA)
     Print(YHA,rsYHA)
     Print(colSums(YHA))
     Print(summary(sumyhat), table(sumyhat5))
   }

   Mrules=NULL

   ######################################################
   ########### Here, sumyhat is ready. ##################
   ######################################################

   if( majestic < 0 ){
     Mrules=named_list( Jall, THall, Valxjall, ValxjRall, Reflect )

     res=list()
     res$j=-1
     res$threshold=NA
     res$nqx=0
     res$locLL=which(sumyhat5 == 1)
     res$locRR=which(sumyhat5 == 0)
     res$left=S[res$locLL]
     res$right=S[res$locRR]
     res$left.score=length(res$locLL)*sum(impfunc(ys[res$locLL]))
     res$right.score=length(res$locRR)*sum(impfunc(ys[res$locRR]))
     res$score=res$left.score + res$right.score
     res$Valxj=NULL
     res$ValxjR=NULL
   }
   else if( majestic %in% c(2,3) ){
     # predict sumyhat5 by X
     res=Divide( sumyhat5, Xs
                 , Ys=NULL, impfunc=impfunc
                 , S=S, method=method, ncatx=ncatx
                 , app=app, maxapp=maxapp, maxncut=maxncut
                 , levelsy=2, levelsX=levelsX )
   }
   else if( majestic %in% c(4,5) ){
     if( majestic == 5 ){
       sumyhat[sumyhat < 0.05]=0.05
       sumyhat[sumyhat > 0.95]=0.95
       sumyhat=log(sumyhat/(1-sumyhat))
     }
     res=Divide( sumyhat, Xs
                 , Ys=NULL, impfunc=variance
                 , S=S, method=method, ncatx=ncatx
                 , app=app, maxapp=maxapp, maxncut=maxncut
                 , levelsy=0, levelsX=levelsX )
   }

   J=res$j

   if( majestic %in% c(3,4,5) ){
    # calculate score in terms of ys
    res$left.score=length(res$locLL)*sum(impfunc(ys[res$locLL]))
    res$right.score=length(res$locRR)*sum(impfunc(ys[res$locRR]))
    res$score=res$left.score + res$right.score
   }
   else if( majestic == 2 ){
    # majestic=2: predict ys from X[,J] in order to find the threshold.
    res=Divide( ys, Xs[,J,drop=0]
               , Ys, impfunc=impfunc
               , S=S, method=method, ncatx=ncatx[J]
               , app=app, maxapp=maxapp, maxncut=maxncut
               , levelsy=levelsy, levelsX=levelsX[J] )
   }

   TH=res$threshold
   k=res$k
   nqx=res$nqx
   left=res$left
   right=res$right
   L=res$left.score
   R=res$right.score
   score=res$score
   Valxj=res$valxj
   ValxjR=res$valxjR
   locLL=res$locLL
   locRR=res$locRR


  } # end of majestic=2 or 3



  info <- list(threshold=TH, j=J, k=k, nqx=nqx
            , left=left, right=right
            , score=score, left.score=L, right.score=R
            , valxj=Valxj, valxjR=ValxjR )
  if( majestic < 0 ) info=append( info, list(Mrules=Mrules) )

  return( info )

} # end of DivideM



