#' Canonical Discriminant Analysis
#'
#' @param y The categorical criterion variable
#' @param X The regressor matrix
#' @param scale = 1 to make the dispersion of the canonical variable I.
#' @param print = 1 to print the result
#'
#' @details
#' This function returns NULL if all the eigenvalues are small.
#'
#' In order to obtain the canonical discriminant variable,
#'  multiply the eigenvectors to X from the right.
#'
#' @examples
#'
#' \dontrun{
#' coef1 <- MASS::lda(Species ~ ., iris)$scaling
#' coef2 <- candisc( iris$Species, iris[,1:4], print=1 )$eve
#' Print(coef1,coef2)
#'
#'
#'
#' # iris: categorical y, categorical X
#' Species <- iris$Species
#' levels(Species) <- c("set","ver","vir")
#' Xc <- iris[,-5]
#' Xc[,3] <- cut(Xc[,3],9, labels=1:9)
#' Xc[,4] <- cut(Xc[,4],9, labels=1:9)
#'
#' y <- Species
#' x <- as.numeric(Xc[,3])
#' temp <- Unique(x)
#' ux <- temp$value
#' loc <- temp$loc
#'
#' # provide numeric design matrix to lda
#' # Xd <- model.matrix(~as.factor(x)-1)
#' # drop the intercept and the first level of x
#' Xd <- model.matrix(~as.factor(x))[,, drop=0][,-1]
#' reslda <- MASS::lda( y~Xd, prior=c(1/3,1/3,1/3) )
#' pred1 <- Xd%*%reslda$scaling
#'
#' eve <- candisc( y, Xd, print=1)$eve
#' pred2 <- Xd%*%eve
#'
#' Print(reslda$scaling,eve)
#' cor(pred1,pred2)
#'
#' }
#'
#' @return
#' A list of: \cr
#' \code{ndim} # of positive eigen values \cr
#' \code{eve} Eigen vector matrix
#'
#' @export
#'

candisc <- function( y, X, scale=0, print=0 ){
  # canonical discriminant analysis or LDA
  # Shin-ichi Mayekawa
  # 20230904,05
  # ndim: 20230914
  # bugfix diag: 20231002
  # matSwp --> Ginv: 20231006
  # return eve=NULL when ndim=0: 20231007cot
  # return ndim: 20231009
  # roxygen2: 20231118
  #

  # Args:
  #  y     a vector of categorical variable
  #  X     regressor matrix
  #
  #
  # Values
  #  eve  the coefficient vector to be multiplied to X from the right as
  #       canvar <- X %*% eve
  #

  nobs=nrow(X)
  nq=ncol(X)
  meant=colMeans(X)

  temp=Unique( y )
  uy=temp$value
  loc=temp$loc

  nG=length(uy)


  B=matrix(0,nq,nq)
  W=B
  meanG=matrix(0,nG,nq)
  for( g in 1:nG ){
    Xg=X[loc[[g]],,drop=0]
    ng=nrow(Xg)
    meang=colMeans(Xg)
    meanG[g,]=meang
    SSCP=meang-t(Xg)
    SSCP=SSCP%*%t(SSCP)
    W=W+SSCP
    B=B+ng*(meang-meant)%*%t(meang-meant)
  }

  if( print >= 2 ){
    Print(W,B)
  }

  # D=B+W
  # DD=mandd(X)$cov*nrow(X)
  # Print(D-DD, fmt="9.7")

  invWB=Ginv(W, MP=1, MASS=1)%*%B
  eig=eigen(invWB)
  eve=Re( eig$vectors )
  eva=Re( eig$values )
  ndim=sum(eva > 1e-9)
  if( ndim > 0 ){
   eve=eve[,1:ndim, drop=0]
   dd=1/sqrt(diag(t(eve)%*%W%*%eve))
   if( scale == 0 ) eve=sqrt(nobs-ndim-1)*eve%*%diag(dd, ncol=ndim)
   else eve=sqrt(nobs)*eve%*%diag(dd, ncol=ndim)
  }
  else eve=NULL


  if( print ){
    Print(meanG)
    Print(scale)
    Print(eva,eve)
    Print(t(eve)%*%eve)
    Print(t(eve)%*%W%*%eve)
  }

  return( list(ndim=ndim,eve=eve) )

} # end of candisc


