#' Cut a Tree
#'
#' This function cuts a tree at the specified depth.\cr
#' Japanese help file: \link[lazy.tree]{cuttree_dt_JPH}
#'
#' @param obj A class DecTree object.
#' @param depth  The depth of the tree at which the tree is cut.
#' @param y The criterion variable vector. (Not Used.)
#' @param nosummary =1 not to calculate summary stat (to speed up)
#' @param nofimport = 1 not to calculate feature importance (to seppe up)
#' @param print = 1 to  print the details.
#' @param plot = 1 to  plot the details.
#' @param title title string
#'
#'
#' @return A \code{DecTree} object.
#'
#' @examples
#' Xc <- data.frame(x1=as.factor(square$x1),x2=as.factor(square$x2))
#' res1 <- DecTree( square$Class4, Xc, minnobs=1, eps=0, method=4, print=3
#'                  , plot=0, maxdepth=4, title="square:" )
#' res2 <- DecTree( square$Class4, Xc, minnobs=1, eps=0, method=4, print=3
#'                  , plot=0, maxdepth=3, title="square:" )
#' rescut <- cuttree_dt( res1, 3, print=1 )
#'
#' Print(res1$mllh, res2$mllh, rescut$mllh)
#' Print(res1$weightedF1, res2$weightedF1, rescut$weightedF1)
#' Print(res2$SCT, "/", rescut$SCT, fmt="5.3")
#'
#' # statname <- c("depth","parent","Left","Right","nobs","score","ImpMeasure")
#' # summary <- get_valLL( rescut$node, statname, simplify=2 )
#' # Print(summary)
#' # plot tree
#' # library(igraph)
#' # plot_dt( rescut )
#'
#' @export
#'

cuttree_dt <- function( obj, depth, y=NULL, nofimport=0, nosummary=0
                        , print=0, plot=0, title="" ){
 # cut tree at given level
 # Shin-ichi Mayekawa
 # 20231015,16
 # DecTree object: 20231016
 # reset j to 0: 20231017dnc
 # names(node)=1 -- nrow(stat): 20231017dnc
 # X as an arg: 20231107dnc
 # plot: 20231107dnc
 # output crosstab: 20231108
 # remove X, use of summary_dt: 20231109
 # roxygen2 and output weightedF1: 20231121dnc
 # Tab -> margin: 20231121dnc
 # nosummary, nofimp: 20231121dnc
 # nlear, aic: 20231123
 # output Node, Gini: 20231125
 # minobs -> minnobs: 20231229
 # MaxDepth: 20240219
 # get_valLL method=2 (some left untouched): 20240523
 # y from obj: 20240523
 #

 # DecTree object
 if( class(obj) == "DecTree" ){
   node=obj$node
   y=obj$y
   MaxDepth=obj$MaxDepth
 }
 else{
    cat("\nerror1:(cuttree_dt) The input object is not a DecTree object.\n")
    stop()
 }

 
 #### below is obsolete since y is from obj.
 if( is.null(y) ){
  cat("\nwarning1:(cuttree_dt) since y is not given, yhat will not be"
      , " calculated.\n")
 }

 # depth info from node list
 depth1=get_valLL( node, "depth", method=2 )
 # Print(depth1)

 if( depth >= max(depth1) ){
  cat("\nerror1:(cuttree_df) The depth parameter is too large.", depth, "\n")
  stop()
 }

 impfuncname=node[[1]]$impfuncname

 # info of depth=depth nodes
 locdeeper=which( depth1 > depth )
 locd=which(depth1 == depth)
 # Print(locd, locdeeper,names(node))
 nodedname=(1:length(node))[locd]
 stat=get_valLL( node[locd], c("depth","Left","Right"), simplify=2, method=1 )
 stat=data.frame(node=nodedname,stat)
 # Print(stat)

 # pick up the middle nodes
 loc=which( stat$Left*stat$Right != 0 )
 stat=stat[loc,]

 # Print(loc)
 # Print(stat)

 # make depth=depth nodes leaf nodes
 for( i in 1:nrow(stat) ){
  nn=stat[i,"node"]
  node[[nn]]$Left=0
  node[[nn]]$Right=0
  node[[nn]]$j=0
  set=node[[nn]]$set
  if( is.null(y) ) ys=NA
  else ys=y[set]
  if( impfuncname %in% c("rss","variance") ){
   node[[nn]]$yhat=mean(ys)
  }
  else{
    node[[nn]]$yhat=Mode(ys)
    prob=table(ys)
    node[[nn]]$prob=prob/sum(prob)
  }
 }

 # remove deeper nodes
 node=node[-locdeeper]

 # renumber the nodes
 stat=get_valLL( node,
   c("depth","parent","Left","Right"), simplify=2, method=1 )
 stat=cbind(node=rownames(stat),stat)
 from=as.numeric(rownames(stat))
 to=1:nrow(stat)
 parent2=recode(stat[,"parent"],from,to)
 Left2=recode(stat[,"Left"],from,to)
 Right2=recode(stat[,"Right"],from,to)
 for( i in 1:nrow(stat) ){
  nn=stat[i,"node"]
  node[[nn]]$parent=parent2[i]
  node[[nn]]$Left=Left2[i]
  node[[nn]]$Right=Right2[i]
 }
 names(node)=1:nrow(stat)   # 20231017dnc
 # Print(node)


 if( print ){
   statname=c("depth","parent","Left","Right","nobs","score","scoreimp"
              , "ImpMeasure")
   stat=get_valLL( node, statname, simplify=2, method=1 )
   stat=data.frame(node=1:nrow(stat), old_node=rownames(stat), stat)
   rownames(stat)=NULL
   Print(stat)
 }

 
 # leaf nodes
 nleaf=sum( stat[,"Left"] == 0 )

 # modify DecTree object
 obj$node=node
 obj$nleaf=nleaf
 obj$MaxDepth=depth
 class(obj)="DecTree"

 fimport=NULL
 if( !nofimport ){
   fimport=feature_imp_dt( obj )
   if( print ) Print(fimport)
   obj$fimport=fimport
   class(obj)="DecTree"
 }
 else{
  obj$fimport=NA
  class(obj)="DecTree"
 }

 if( !is.null(y) && !nosummary ){
   temp=summary_dt( obj, print=print, outpred=1 )
   obj$Tab=temp$Tab
   obj$margin=temp$margin
   obj$corr=temp$corr
   obj$Gini=temp$Gini
   obj$mllh=temp$mllh
   obj$nparam=temp$nparam
   obj$aic=temp$aic
   obj$bic=temp$bic
   obj$tprecision=temp$tprecision
   obj$macroF1=temp$macroF1
   obj$weightedF1=temp$weightedF1
   obj$SCT=temp$SCT
   obj$yhat=temp$yhat
   obj$y=y
   obj$prob=temp$prob
   obj$Node=temp$Node
   class(obj)="DecTree"
 }
 else{
   obj$Tab=NA
   obj$margin=NA
   obj$corr=NA
   obj$Gini=NA
   obj$mllh=NA
   obj$nparam=NA
   obj$aic=NA
   obj$bic=NA
   obj$tprecision=NA
   obj$macroF1=NA
   obj$weightedF1=NA
   obj$SCT=NA
   obj$yhat=NA
   obj$prob=NA
   obj$Node=NA
   class(obj)="DecTree"

 }


 if( plot ){
  plot_dt( obj )
 }

 return( obj )

} # end of cuttree_dt




