#' Display the Definition of the Node
#'
#' This function displays the definition of the specified nodes.\cr
#' Japanese help file: \link[lazy.tree]{node_info_dt_JPH}
#'
#' @param obj The output of DecTree or list of nodes.
#' @param depth The depth (scalar) \cr
#' To display all the leaf nodes, use \code{depth > 9999}.
#' @param node Vector of the node number \cr
#' This has priority over \code{depth}.
#' @param nohist = 1 not to calculate division history or node definition.
#' @param simplify = 0 NOT to simplify the continuous regressor values. \cr
#' (Opt for this option when you notice something unusual about the results.)
#' @param title The title string
#' @param print = 1 to output node_info
#' @param debug = 1 to print intermediate info
#'
#'
#'
#' @examples
#' # categorical y, continuous X1 and categorical X2
#' plot(square$x1,square$x2, type="n")
#' text(square$x1,square$x2,square$Class4, cex=2)
#' SQ2=matrix(square$class4, 4,, byrow=1)
#' heatmap( t(SQ2), Rowv=NA, Colv=NA, reorderfun=NA, symm=1
#'        , col=c("darkred","darkgreen","orange","lightblue")
#'        , labRow=4:1, labCol=1:4, revC=TRUE )
#' legend( "left", legend=1:4, pch=1,cex=2, pt.lwd=2, box.lwd=2
#'         , col=c("darkred","darkgreen","orange","lightblue") )
#' Xc <- data.frame(x1=square$x1,x2=square$c2)
#' yc <- as.factor(square$class4)
#'
#' # specify plot=1 or use plot_dt if the package igraph is installed.
#' res0 <- DecTree( yc, Xc, maxdepth=4, method=4, print=1, impfuncname="gini"
#'                  , plot=0, title="square: numerical X1, categorical X2" )
#' # plot_dt(res0, col_leaf=c("darkred","darkgreen","orange","lightblue","red")
#' #       , title="square: numerical X1, categorical X2" )
#'
#' # nodes at depth 2
#' temp <- node_info_dt( res0, depth=2, title="depth" )
#' # nodes 4, 15, and 24
#' temp <- node_info_dt( res0, node=c(8,19,21), title="node" )
#'
#' # leaf node info of the above tree
#' info0 <- node_info_dt( res0, depth=99, print=2, title="num X1 and cat X2" )
#'
#' res1 <- DecTree( yc, data.frame(x1=square$x1,x2=square$x2)
#'                  , maxdepth=4, method=4, print=1, impfuncname="gini"
#'                  , plot=0, title="square: numeric X1 and X2" )
#' # plot_dt( res1, col_leaf=c("darkred","darkgreen","orange","lightblue","red")
#' #       , title="square: numeric X1 and X2" )
#' info1 <- node_info_dt( res1, depth=99, print=2, title="num X1 and num X2" )
#'
#' res2 <- DecTree( yc, data.frame(x1=square$c1,x2=square$c2)
#'                  , maxdepth=4, method=4, print=1, impfuncname="gini"
#'                  , plot=0, title="square: categorical X1 and X2" )
#' # plot_dt( res2, col_leaf=c("darkred","darkgreen","orange","lightblue","red")
#' #       , title="square: categorical X1 and X2" )
#' info2 <- node_info_dt( res2, depth=99, print=2, title="cat X1 and cat X2" )
#'
#'
#' @return A list of \cr
#'
#' \code{NodeInfo}
#' \preformatted{
#' Node       the node number.
#' PA         the parent number.
#' N          number of observations in the node.
#' Score      weighted impurity measure of the node.
#' ImpMeasure impurity measure of the node (score divided by # of obs)
#' SplitVar   the splitting variable number of this node
#'            to be used to divide THIS node.
#' LR         = "L" if the node is the left node of the parent.
#' Leaf       = 1 if the node is a leaf node.
#' Yhat       the predicted value at the leaf nodes
#' Prob       the predicted probability at the leaf nodes
#' }
#'
#' \code{NodeHist} A list of division history of each node.
#' \preformatted{
#' Depth  the depth level
#' Parent the parent of the leaf at depth Depth
#' J      the regressor variable number used to divide Parent
#' Jn     the name of the regressor J
#' Ncx    the # of levels of regressor J
#' LR     = "L" if this leaf node is classified as Left
#' SW     = 1 if the Left and the Right were switched
#' Score  obs. weighted impurity score
#' ImpMesure impurity score (divided by nobs)
#' Valxj._v_1, Valxj._v_2, ...  The value of the categorical regressor J
#' }
#'
#' and \cr
#' \code{NodeDef} A list of node definition of each node.
#'
#' @export
#'


node_info_dt <- function( obj, depth=99999, node=NULL, nohist=0
                        , title="", simplify=1, print=1, debug=0 ){
  # print leaf impurity
  # Shin-ichi Mayekawa
  # 20230910
  # node --> Node, vertex --> node: 20230928
  # print_leaf --> print_leaf_dt: 20231002
  # DecTree object: 20231016
  # Prob: 20231024mm
  # dump: 20231024mm
  # leafinfo: 20240209,10,11,13,14,15,16
  # print_leaf_dt renamed as leaf_info_dt: 20240216
  # regval not sorted: 20240216
  # levelsX: 20240216dnc,16
  # simplify: 20240216,17
  # seq_len by chatGPT: 20240217
  # node_info: 20240218,19
  # new simp_ineq: 20240220,21
  # RegVal -> NodeDef, regval -> NodeHist, leaf_info -> node_info: 20240222
  # output ImpurityMeasure: 20240224,25,26
  # leafinfo --> nodeinfo,  lh --> nih: 20240226
  # del: 20241226
  # output prob in NodeInfo: 20240303
  # output splitting variable and LR: 20240304,07
  # roxgen2: 20240306dnc
  # del Valxj_v_1 --> _v_1: 20240327
  # debug: 20240330
  # get_valLL -> sapply: 20240401
  # majestic<0: 20240417
  # bugfix for continuous y, get_valLL depth: 20240518
  # get_valLL method=2: 20240523
  # check if depth is a scalar: 20240523
  # get_valLL valxj cannot method=2: 20240524
  # top node: 20240607,08
  # 

  # error
  if( length(depth) > 1 ){
   cat("error1:(node_info_dt) depth argument must be a scalar.\n")
   stop()
  }
  
  if( length(node) > 1  &&  1 %in% node ){
   cat("error1:(node_info_dt) node=1 cannot be specified"
   , " when node is a vector.\n")
   stop()
  }
  
  # DecTree object
  if( class(obj) == "DecTree" ){
   ncaty=obj$ncaty
   impfuncname=obj$impfuncname
   ncatx=obj$ncatx
   xname=obj$xname
   levelsX=obj$levelsX
   levelsy=obj$levelsy
   MaxDepth=obj$MaxDepth
   majestic=obj$majestic
  }
  else if( is.list(obj) & !is.data.frame(obj) ){
    nam=names(obj[[1]])
    if( all(
    c("parent","set","depth","nobs","Left","Right","j") %in% names(obj[[1]])
    ) ){
      lastnode=obj[[length(obj)]]
      if( all(is.na(lastnode$prob)) ){
       ncaty=0
       ylevel=NULL
      }
      else{
       ncaty=length(lastnode$prob)
       ylevel=names(lastnode$prob)
      }
      obj=list(node=obj, ylevel=ylevel)
      }
    impfuncname=NA
    ncatx=NA
    xname=NA
    MaxDepth=max(get_valLL(obj$node,"depth", method=2))
    levelsX=NA
    levelsy=1:ncaty
  }
  else{
    cat("\nerror1:(node_info_dt) The input object is not a DecTree object.\n")
    stop()
  }

  # node: A list of node information where node[[r]]  contains:
  #  node[[r]]$parent   The parent of the r-th node
  #  node[[r]]$set      The set of observations in the r-th node
  #  node[[r]]$j        The variable to be used to divide this node
  #  node[[r]]$k     *  The quantified variable number used.
  #  node[[r]]$th    *  The value of the threshold.
  #  node[[r]]$nqx   *  # of unique quantified variables.
  #  node[[r]]$valxj *  The set of factor levels of categorical variable j
  #                       to be classified as left from this node.
  #  node[[r]]$valxjR *  The set of factor levels of categorical variable j
  #                       to be classified as right from this node.
  #  node[[r]]$scoreimp The improvement of impurity measure due to dividing
  #                       this node.
  #  node[[r]]$depth    The distance from the top node
  #  node[[r]]$Left  *  The left node name
  #  node[[r]]$Right *  The right node name
  #
  #  node[[r]]$yhat  ** The predicted value
  #  node[[r]]$prob  ** The post probability vector
  #
  #  * Not included in the final nodes.
  #  ** Not included in the middle nodes.

  # # of nodes and their name
  m=length(obj$node)
  nodename=names(obj$node)

  # error
  if( length(depth) > 1 ){
   cat("error1:(node_info_dt): depth must be a scalar.", depth,"\n")
   stop()
  }
  if( is.null(node)  &&  depth < 0 ){
   cat("error1:(node_info_dt): depth out of range.", depth,"\n")
   stop()
  }
  if( !is.null(node) ){
    if( any(node <= 0)  ||  any(node > m) ){
      cat("error1:(node_info_dt): node out of range.", node,"\n")
      stop()
    }
  }

  Node=NULL
  PA=NULL
  SplitVar=NULL
  ImpMeasure=NULL
  Score=NULL
  N=NULL
  Leaf=NULL
  LR=NULL
  Yhat=NULL
  if( ncaty > 0 ){
   Prob=matrix(0,length(obj$node),ncaty)             # When contimuous!!!!!!!!!
   colnames(Prob)=paste("ycat",levelsy,sep="_")
  }
  else{
   Prob=NA
  }
  
  stop=0

  # for each leaf node, store the node info
  if( !is.null(node) ){
    # node number given
    for( h in node ){
        Node=c(Node,h)
        pa=obj$node[[h]]$parent
        PA=c(PA,pa)
        if( pa == 0 ){
         LR=c(LR,"root")
        }
        else if( h == obj$node[[pa]]$Left ) LR=c(LR,"L")
        else LR=c(LR,"R")
        ImpMeasure=c(ImpMeasure, obj$node[[h]]$ImpMeasure)
        Score=c(Score, obj$node[[h]]$score)
        N=c(N,length(obj$node[[h]]$set))
        Leaf=c(Leaf, as.numeric(obj$node[[h]]$j == 0))
        SplitVar=c(SplitVar,obj$node[[h]]$j)
        #      Yhat=c(Yhat, obj$node[[h]]$yhat)
        #      if( ncaty > 0 ) Prob=rbind(Prob,obj$node[[h]]$prob)
    }
   if( all(node == 1) ) stop=1
  }
  else if( depth <= MaxDepth ){
    # node depth given
    for( h in 1:length(obj$node )){
      if( obj$node[[h]]$depth == depth ){
        Node=c(Node,h)
        pa=obj$node[[h]]$parent
        PA=c(PA,pa)
        if( pa == 0 ){
         LR=c(LR,"root")
        }
        else if( h == obj$node[[pa]]$Left ) LR=c(LR,"L")
        else LR=c(LR,"R")
        SplitVar=c(SplitVar,obj$node[[h]]$j)
        ImpMeasure=c(ImpMeasure, obj$node[[h]]$ImpMeasure)
        Score=c(Score, obj$node[[h]]$score)
        N=c(N,length(obj$node[[h]]$set))
        Leaf=c(Leaf, as.numeric(obj$node[[h]]$j == 0))
        #      Yhat=c(Yhat, obj$node[[h]]$yhat)
        #      if( ncaty > 0 ) Prob=rbind(Prob,obj$node[[h]]$prob)
      }
    }
   if( depth == 0 ) stop=1
  }
  else{
    # leaf node
    for(h in 1:length(obj$node)){
      if( obj$node[[h]]$j == 0 ){
        Node=c(Node,h)
        pa=obj$node[[h]]$parent
        PA=c(PA,pa)
        if( h == obj$node[[pa]]$Left ) LR=c(LR,"L")
        else LR=c(LR,"R")
        SplitVar=c(SplitVar,obj$node[[h]]$j)
        Yhat=c(Yhat, obj$node[[h]]$yhat)
        ImpMeasure=c(ImpMeasure, obj$node[[h]]$ImpMeasure)
        Score=c(Score, obj$node[[h]]$score)
        N=c(N,length(obj$node[[h]]$set))
        if( ncaty > 0 ){
         Prob[h,]=obj$node[[h]]$prob
         Prob=rbind(Prob,obj$node[[h]]$prob)
        }
      }
    }
    nleaf=length(Node)
    Leaf=rep(1,nleaf)
    if( ncaty > 0 ){
     Prob=Prob[Node,,drop=0]
     colnames(Prob)=paste("ycat",levelsy,sep="_")
    }
  }

  nnode=length(Node)

  Yhat[Leaf==0]=NA
  if( ncaty > 0 ) Prob=Prob[1:nnode,,drop=0]
  if( any(Leaf==0) ) Prob[which(Leaf==0),]=NA
  if( print > 0 ){
   cat("\nNode Information: ", title, "\n", sep="")
   if( is.null(node) ){
     if( depth <= MaxDepth ) cat("\n # of nodes at depth", depth, " =", nnode)
     else    cat("\n # of Leaf Nodes =", nleaf)
   }
   else cat("\n # of nodes to be inspected =", nnode)
   cat("\n Leaf Impurity (",impfuncname,")")
   cat("\n PA is the name of the parent node.")
   cat("\n N is the # of observations in the node.")
   cat("\n Leaf = 1 if the node is a leaf node.")
   cat("\n ImpMeasure is theimpurity measure of the node.\n")
   if( ncaty > 0 )
    printm(Node, PA, LR, N, ImpMeasure, Score, Leaf, SplitVar, Yhat, Prob )
   else
    printm(Node, PA, LR, N, ImpMeasure, Score, Leaf, SplitVar, Yhat )
   # printm(Node, PA, N, ImpMeasure, Score, Leaf )
  }

  # to be returned at the end
  NodeInfo=data.frame(Node, PA, LR, N, Score, ImpMeasure, Leaf
                      , SplitVar, Yhat, Prob)


  # return only the NodeInfo.
  if( stop  ||  nohist ){
    return( NodeInfo )
  }


  if( majestic < 0 ){
    cat("\nwarning1:(node_info_dt) Node definition cannot be calculted")
    cat(" when majestic < 0.\n")
    return( NodeInfo )
  }


  Score0=Score; ImpMeasure0=ImpMeasure

  # for each leaf node, store the regressor characteristics
  nodeinfo=list()
  for( h in 1:nnode ){

   nhall=list()
   pah=PA[h]
   nodeh=Node[h]
   paall=list()
   # back track to the top node
   while( pah >= 1 ){
    paall=append(paall,list(pah))
    nh=obj$node[[pah]][
    c("j","th","valxj","valxjR","Left","Right","switched","score","ImpMeasure")
    ]
    nhall=append(nhall,list( list(nodeh=nodeh,pa=pah,nh=nh)) )
    pah=obj$node[[pah]]$pa
   }
   names(nhall)=unlist(paall)
   nodeinfo[[h]]=nhall
#   str(nhall,2)

  } # end of h
  names(nodeinfo)=Node

  if( debug > 0 ){
    cat("\n*****nodeinfo*******\n")
    Print(nodeinfo)
  }

# cat("after nodeinfo\n")
# etime(2)

  # for each node, find the values of the regressors
  # and reverse the order
  cond=list()
  for( h in 1:nnode ){
   
    if( Node[h] == 1 ) next
   
    nih=nodeinfo[[h]]

    condp=list()
    currentnode=Node[h]
    lenlh=length(nih)

    # back track to the top node
    for( p in 1:length(nih) ){

      if( currentnode == nih[[p]]$nh$Left ){
        LR="L"
        valxj=nih[[p]]$nh$valxj
      }
      else{
        LR="R"
        valxj=nih[[p]]$nh$valxjR
      }

      condp[[lenlh+1-p]]=list( j=nih[[p]]$nh$j,th=nih[[p]]$nh$th
                , LR=LR, valxj=valxj, switched=nih[[p]]$nh$switched
                , score=nih[[p]]$nh$score, ImpMeasure=nih[[p]]$nh$ImpMeasure)
      currentnode=nih[[p]]$pa

    } # end of parents

    names(condp)=names(nih)[rev(1:lenlh)]
    cond[[h]]=condp

  } # end of h for leaf node

  names(cond)=Node

  if( debug > 0 ){
    cat("\n******** cond ************\n")
    Print(cond)
  }

# cat("after cond\n")
# etime(2)

  #
  #
  # Above two loops indexed by h can be combined.
  #
  #

  # for each node, find the values of the regressors
  # and reverse the order
  NodeHist=list()
  for( h in 1:nnode ){
   
   if( Node[h] == 1 ) next
   
   condh=cond[[h]]
   # J=get_valLL(condh,"j", removeNull=1)
   J=sapply(condh, function(x) x[["j"]])
   Jn=xname[J]
   # TH=get_valLL(condh,"th", removeNull=1)
   TH=sapply(condh, function(x) x[["th"]])
   # LR=get_valLL(condh,"LR", removeNull=1)
   LR=sapply(condh, function(x) x[["LR"]])
   # SW=get_valLL(condh,"switched", removeNull=1)
   SW=sapply(condh, function(x) x[["switched"]])
   # Score=get_valLL(condh,"score", removeNull=1)
   Score=sapply(condh, function(x) x[["score"]])
   # ImpMeasure=get_valLL(condh,"ImpMeasure", removeNull=1)
   ImpMeasure=sapply(condh, function(x) x[["ImpMeasure"]])
   colnames(J)=colnames(TH)=colnames(LR)=colnames(SW)=
     colnames(Score)=colnames(ImpMeasure)=NULL
 #  Valxj0=get_valLL(condh,"valxj", removeNull=1)
   Valxj=get_valLL(condh,"valxj", method=1, debug=0 )

   nn=Node[h]
   Depth=(1:length(J))-1

   ch=data.frame( Depth=Depth
                , Parent=names(condh),J=J,Jn=Jn, Ncx=ncatx[J],TH=TH,LR=LR,SW=SW
                , Score, ImpMeasure, Valxj=Valxj )
   rownames(ch)=NULL

   NodeHist[[h]]=cbind(ch)

  } # end of h for leaf node

  names(NodeHist)=Node

  if( debug > 0 ){
    cat("\n******** NodeHist ************\n")
    Print(NodeHist)
  }

# cat("after NodeHist\n")
# etime(2)

  varnum=c("Depth", "Parent", "J", "Jn", "Ncx", "TH", "LR", "SW")
  varcatm=c("LR", "SW")

  # divide NodeHist into regnum and regcat
  regnum=list()
  regcat=list()
  for( h in 1:nnode ){
   
    if( Node[h] == 1 ) next
   
    condh=NodeHist[[h]]

    # sort by regressor
    od=order(condh$J)
    condh=condh[od,]
    rownames(condh)=NULL


    locnum=which(condh$Ncx == 0)
    loccat=setdiff(1:nrow(condh),locnum)
    if( length(locnum) > 0 ) regnum[[h]]=condh[locnum,varnum, drop=0]
    else regnum[[h]]=matrix(NA,1,1)
    if( length(loccat) > 0 ) regcat[[h]]=condh[loccat,, drop=0]
    else regcat[[h]]=matrix(NA,1,1)

  } # end of h for leaf node


  names(regnum)=names(regcat)=Node

  if( debug > 0 ){
    cat("\n******** regnum ************\n")
    Print(regnum,regcat, nnode)
  }

# cat("after regnum\n")
# etime(2)

  # combinations of LR and SW
  # If SW==0,
  #   LR="L" left  node: xj <= th
  #   LR="R" right node: xj >  th
  # If SW==1,
  #   LR="L" left  node: xj >  th
  #   LR="R" right node: xj <= th
  #
  Oper=matrix(c("<=",">",">","<="), 2,2,byrow=1)
  rownames(Oper)=c("L","R")

  NodeDef=list()
  # final values
  for( h in 1:nnode ){
   
    if( Node[h] == 1 ) next
   
    rnumh=regnum[[h]]
    rcath=regcat[[h]]

    Cnum=NULL
    Cnuml=NULL
    Ccat=NULL


    # num  when rnumh != matrix(NA,1,1)
    # Convert the condition to:  variable J inequality-sign  threshold
    # Note that variable name is placed to the left of the inequality sign.
    if( ncol(rnumh) > 1 ){

      Jh=rnumh$J
      uJh=unique(Jh)
      nuJh=length(uJh)
      #Print(h,rnumh,Jh,nuJh)

      Op=Oper[rnumh[1,"LR"],rnumh[1,"SW"]+1]
      for( u in 1:nuJh ){
        loc=which(Jh==uJh[u])
        Cnumi=NULL
        for( i in 1:length(loc) ){
          Op=Oper[rnumh[loc[i],"LR"],rnumh[loc[i],"SW"]+1]
          txt=paste(rnumh[loc[i],"Jn"],Op,rnumh[loc[i],"TH"], sep=" ")
          if( i == 1 ){
            Cnumi=txt
          }
          else{
            Cnumi=paste(Cnumi,txt, sep=" & ")
          }
          temp=simp_ineq( Cnumi, debug=debug )
          if( simplify == 1 ){
             Cnumi=temp$condc
             condl=temp$condl
          }
        } # end of i

        if( debug > 0 ) Print(h,u,i,Cnumi,condl)

        Cnum=paste(Cnum,Cnumi,sep=" & ")
        Cnuml=rbind(Cnuml,condl)
      } # end of u

      if( length(Cnum) > 0 && substring(Cnum,1,3) == " & " )
        Cnum=substring(Cnum,4)

    }


    # cat
    if( length(rcath) > 1 ){
      del=which( regexpr("_v_1",colnames(rcath)) > 0 )[1]-1
      uJ=unique(rcath$J)
      nuJ=length(uJ)
      for( i in 1:nuJ ){
       loc=which(rcath$J == uJ[i])
       ########  Depth Parent J Jn Ncx (TH) LR SW Score ImpMeasure  #######
       vxj=rcath[loc[length(loc)],-c(1:del)]  # remove Depth--ImpMeasure.
       vxj=vxj[!is.na(vxj)]
#  Print(uJ[i], vxj,levelsX[[uJ[i]]],levelsX[[uJ[i]]][as.numeric(vxj)])
       vxj=levelsX[[uJ[i]]][as.numeric(vxj)]
       nj=rcath[loc[length(loc)],"Jn"]
       txt=paste(nj,"=(", paste(vxj, collapse=","),")", sep="")
       Ccat=paste(Ccat,txt, sep=" & ")
      }
    }

    if( length(Ccat) > 0  &&  substring(Ccat,1,3) == " & " )
      Ccat=substring(Ccat,4)

    impmeasure=c(Score=Score0[h], ImpMeasure=ImpMeasure0[h])

    NodeDef=append( NodeDef,list( list(Cnum=Cnum, Cnuml=Cnuml, Ccat=Ccat
                          , impmeasure=impmeasure ) ) )


  } # end of h loop

  names(NodeDef)=Node[Node != 1]

# cat("after NodeDef\n")
# etime(2)

  if( length(Yhat) == length(NodeDef) ) is.leaf=1
  else is.leaf=0

  # Print(length(Yhat),length(NodeDef))
Print(NodeDef, ImpMeasure0)
  if( print >= 1 ){
    cat("\nNode Definition\n")
    if( !is.leaf ){
      for( i in 1:length(NodeDef) ){
        cat("\n\nnode", names(NodeDef)[i], ",  N =", N[i]
            , ": ImpMeasure =", impmeasure[i],"\n")
        cat(" ", NodeDef[[i]]$Cnum, "\n")
        cat(" ", NodeDef[[i]]$Ccat)
      }
    }
    else{
      for( i in 1:length(NodeDef) ){
        cat("\n\nnode", names(NodeDef)[i], ": class =", Yhat[i]
            , ",  N =", N[i]
            , ",  ImpMeasure =", ImpMeasure0[i],"\n")
        cat(" ", NodeDef[[i]]$Cnum, "\n")
        cat(" ", NodeDef[[i]]$Ccat)
      }

    }
  }

  res=named_list(NodeInfo, NodeHist, NodeDef)

  return( res )

 } # end of node_info_dt



