#' Plot Decision Tree
#'
#' @param obj The output of DecTree or list of nodes.
#' @param title The title string
#' @param type = 2 for numeric criterion
#' @param xname Vector consisting of the names of the regressors.\cr
#' This is to provide the regressor names when obj is just a list of nodes.
#' @param col_leaf Color names for the leaf nodes, or NULL
#' @param pch point type of legend
#' @param lwd point size of legend
#' @param cex character size of legend
#'
#' @details
#' This function requires \code{igraph} package.
#'
#' @return A dataframe consisting of
#' \preformatted{
#' nodename = node name
#' nodenum = leaf node name
#' varnum = variable name
#' Col = color of the nodes#'
#' }
#'
#'
#' @examples
#' \dontrun{
#' library(igraph)
#' Xc <- as.data.frame(cbind(square$c1,square$c2))
#' res1 <- DecTree( square$Class4, Xc, maxdepth=3, print=1 )
#' plot_dt( res1, title="square"
#'        , col_leaf=c("red","darkgreen","orange","lightblue") )
#' plot_dt( res1$node, title="square", xname=res1$xname )
#' }
#'
#'
#' @export
#'

plot_dt <- function( obj, title="Decision Tree", type=1, xname=NULL
                     , col_leaf=NULL, pch=1, lwd=4, cex=0.7){
 # plot result
 # Shin-ichi Mayekawa
 # 20230907
 # layout_as_tree: 20230920
 # plot type=2: 20230927
 # vertex --> node: 20230928
 # DecTree object: 20231016,17
 # out of DecTree31.R  and roxygen2: 20231208dnc
 # new color: 20240110
 # col_leaf: 20240224,26
 # pch etc: 20240306dnc
 # majestic<0: 20240417
 # get_valLL method=2: 20240523
 #

 # preparation for igraph
 library(igraph)

 # DecTree object
 if( class(obj) == "DecTree" ){
  node=obj$node
  xname=obj$xname
  rm(obj)
 }
 else if( is.list(obj) & !is.data.frame(obj) ){
  nam=names(obj[[1]])
  if( all(
   c("parent","set","depth","nobs","Left","Right","j") %in% names(obj[[1]])
  ) ){
   node=obj
   rm(obj)
  }
 }
 else{
  cat("\nerror1:(plot_dt) The input object is not a DecTree object.\n")
  stop()
 }

 m=length(node)

 #col <- array(dim=m)
 col=numeric(m)
 edge.list <- matrix(nrow=m, ncol=2)
 for (h in 1:m)
  edge.list[h, ] <- c(node[[h]]$parent, h)
 edge.list <- edge.list[-1, ]
 colnames(edge.list)=c("parent", "node")
 for (h in 1:m)
  col[h] <- node[[h]]$j
 col1=col
 col[col==-1]=1

 if( 0 ){
   # igraph
   g <- graph_from_edgelist(edge.list)
   V(g)$color <- col
   # V(g)$name <- ""
   plot(g, pin=c(4, 4), layout=layout_as_tree(g, root=1))
   t1=paste(title, ": colored by regressor")
   title( t1 )
 }


 # correct node colors
 vertex_name=c( get_valLL(node,"yhat", removeNull=1, method=2) )
 if( length(vertex_name) == 0 ) return( invisible() )

 isnum=which( is.number( lapply(vertex_name, as.list), pat=2 ) )
 if( type == 2 ){
   vertex_name[isnum]=
     formatC(as.numeric(vertex_name[isnum]),digits=2, format="g")
 }
 vertex_num=recode(vertex_name,unique(vertex_name)
                   , 1:length(unique(vertex_name)))
 vnum=numeric(length(node))
 vnum[!is.na(vertex_num)]=as.numeric(vertex_num[!is.na(vertex_num)])-1
 vertex_name[is.na(vertex_name)]=(1:m)[is.na(vertex_name)]



 ColorCool=c(
   "lightsteelblue3", "aquamarine4", "darkslategray1", "lightslateblue"
   , "aquamarine", "limegreen", "deepskyblue4", "cornflowerblue"
   , "deepskyblue", "dodgerblue4", "darkslateblue", "mediumblue"
   , "navy", "royalblue4", "slategrey", "gray30", "gray47", "gray"
   , "gray70", "gray88"
 )
 ColorWarm=c(
   "lightpink", "lightsalmon", "orange3", "sienna1", "palegoldenrod"
   , "salmon2", "salmon4", "pink", "pink4", "palevioletred"
   , "sienna4", "tomato", "thistle1", "wheat", "red", "red4", "yellow"
   , "yellow2", "yellowgreen", "black"
 )

 Col=character(m)

 col[col>0]=(col[col>0]-1) %% length(ColorCool) + 1
 Col[col>0]=ColorCool[col[col>0]]

 if( is.null(col_leaf) ){
  vnum[vnum>0]=(vnum[vnum>0]-1) %% length(ColorWarm) + 1
  Col[vnum>0]=ColorWarm[vnum[vnum>0]]
 }
 else{
  Col[vnum>0]=col_leaf[as.numeric(vertex_name[vnum>0])]
 }

  # Print(vertex_name,vnum, col, Col)
  # Print(xname,vertex_name[vnum>0])

 leg=character(m)
 if( !is.null(xname) ) leg[col>0]=xname[col[col>0]]
 else leg[col>0]=col[col>0]
 leg[vnum>0]=vertex_name[vnum>0]

 leg=leg[col1>0]
 leg=paste((1:length(col1))[col1>0],leg,sep=":")
 legc=Col[col1>0]

 g <- graph_from_edgelist(edge.list)
 V(g)$color <- Col
 V(g)$name <- vertex_name
 resplot=plot(g, pin=c(4, 4), layout=layout_as_tree(g, root=1))
 t1=paste(title)
 title( t1 )
 if( length(leg) > 0 )
  legend("topleft", legend=leg, pch=pch, lty=0, col=legc, lwd=lwd, cex=cex)

 return( invisible(
   data.frame(nodename=vertex_name, nodenum=vnum, varnum=col, Col) ) )

} # end of plot_dt


