#' Print Division History
#'
#' @param obj The output of DecTree or list of nodes.
#' @param title The title string
#' @param dump = 1 to output result
#'
#' @examples
#' Xc <- data.frame(square$c1,square$c2)
#' res1 <- DecTree( square$Class4, Xc, maxdepth=4, print=0 )
#' print_hist_dt( res1$node, title="square" )
#'
#' @return A list of a data frame \code{hist} and \code{Valxj} \cr
#' where data frame \code{hist} consists of
#' \preformatted{
#' Node the node number
#' PA the parent node number
#' VAR the regressor variable number to be used to divide the node
#' k the quantified variable number to be used
#' TH the threshold to be used
#' Valxc compact version of Valxj below.
#' }
#' and \code{Valxj} is a list of the values categorized as "LEFT"
#' among the levels of categorical regressors used when splitting each node.
#'
#'
#' @export
#'
print_hist_dt <- function( obj, title="", dump=0 ){
  # print tree history with categorical regressor values
  # Shin-ichi Mayekawa
  # 20230906
  # always output threshold: 20230928
  # node --> Node, vertex --> node: 20230928
  # print_hist --> print_hist_dt: 20231002
  # long Varxc: 20231007cot
  # DecTree object: 20231016
  # dump: 20231024mm
  # title: 20231110cot
  # cat strings corrected, dump Valxj: 20231219dnc
  #

  # DecTree object
  if( class(obj) == "DecTree" ){
    node=obj$node
    rm(obj)
  }
  else if( is.list(obj) & !is.data.frame(obj) ){
    nam=names(obj[[1]])
    if( all(
    c("parent","set","depth","nobs","Left","Right","j") %in% names(obj[[1]])
    ) ){
      node=obj
      rm(obj)
    }
  }
  else{
   cat("\nerror1:(print_hist_dt) The input object is not a DecTree object.\n")
    stop()
  }

  VAR <- NULL
  TH <- NULL
  PA <- NULL
  Node=NULL
  Valxj=NULL
  k=NULL
  for (h in 1:length(node)) {
    if (node[[h]]$j != 0) {
      j <- node[[h]]$j
      valxj <- node[[h]]$valxj
      th <- node[[h]]$th
 #     if( !is.null(valxj)) th=NA

      VAR <- c(VAR, j)
      k=c(k,node[[h]]$k)
      TH <- c(TH, th)
      Node <- c(Node, h)
      PA=c(PA,node[[h]]$parent)
      Valxj <- append(Valxj, list(valxj))
    }
  }
  names(Valxj)=Node
  Valxc=unlist( lapply( Valxj, paste, collapse=", ") )
  cat("\n", title, sep="")
  cat("\nDivision History")
  cat("\n PA is the name of the parent node.")
  cat("\n VAR is the variable used to divide the node.")
  cat("\n k is the number of the quantified j-th regressor variable used.")
  cat("\n  k=NA means app=1 was used.")
  cat(
   "\n TH is the threshold for division when VAR is continuous or quantified.")
  cat("\n  Those with VAR <= TH were classified to the left.")
  cat("\n When VAR is categorical, Valxc contains the category values of VAR")
  cat("\n  to be classified to the left.")

  # cat("\n node,       parent,       variable, thresold,   categories\n")
  if( max(nchar(Valxc)) < 10 ){
   printm(Node, PA, VAR, k, TH, Valxc)
  }
  else{
   printm(Node, PA, VAR, k, TH)
   printm(Node,Valxc)
  }

  if( dump == 1 ){
   hist=data.frame( Node, PA, VAR, k, TH, Valxc)
   res=list(hist=hist, Valxj=Valxj)
   return( res )
  }

} # end of print_hist_dt


