#' Print the Leaf Node Information
#'
#' @param obj The output of DecTree or list of nodes.
#' @param title The title string
#' @param dump = 1 to output result
#'
#' @examples
#' Xc <- data.frame(square$c1,square$c2)
#' res1 <- DecTree( square$Class4, Xc, maxdepth=4, print=0 )
#' print_leaf_dt( res1$node, title="square" )
#'
#' @return A data frame consisting of
#' \preformatted{
#' Node the node number
#' PA   the parent number
#' N    number of observations in the node
#' ImpMeasure impurity measure of the node
#' Yhat the predicted value
#' Prob the probability of each level of the categorical criterion
#' (column names of Prob matrix will be used as the data frame variable names.)
#' }
#'
#'
#' @export
#'
print_leaf_dt <- function( obj, title="", dump=0 ){
  # print leaf impurity
  # Shin-ichi Mayekawa
  # 20230910
  # node --> Node, vertex --> node: 20230928
  # print_leaf --> print_leaf_dt: 20231002
  # DecTree object: 20231016
  # Prob: 20231024mm
  # dump: 20231024mm
  #

  # DecTree object
  if( class(obj) == "DecTree" ){
   ncaty=obj$ncaty
  }
  else if( is.list(obj) & !is.data.frame(obj) ){
    nam=names(obj[[1]])
    if( all(
    c("parent","set","depth","nobs","Left","Right","j") %in% names(obj[[1]])
    ) ){
      lastnode=obj[[length(obj)]]
      if( all(is.na(lastnode$prob)) ){
       ncaty=0
       ylevel=NULL
      }
      else{
       ncaty=length(lastnode$prob)
       ylevel=names(lastnode$prob)
      }
      obj=list(node=obj, ylevel=ylevel)
       }
  }
  else{
    cat("\nerror1:(print_leaf_dt) The input object is not a DecTree object.\n")
    stop()
  }

  m=length(obj$node)

  Node=NULL
  PA=NULL
  ImpMeasure=NULL
  Yhat=NULL
  N=NULL
  Prob=NULL
  for(h in 1:length(obj$node)){
    if( obj$node[[h]]$j == 0 ){
      Node=c(Node,h)
      PA=c(PA,obj$node[[h]]$parent)
      Yhat=c(Yhat, obj$node[[h]]$yhat)
      ImpMeasure=c(ImpMeasure, obj$node[[h]]$ImpMeasure)
      N=c(N,length(obj$node[[h]]$set))
      if( ncaty > 0 ) Prob=rbind(Prob,obj$node[[h]]$prob)
    }
  }

  colnames(Prob)=obj$ylevel
  #colnames(Prob)=rep("",ncol(Prob))
  cat("\n", title, "\n", sep="")
  cat("\nLeaf Impurity\n")
  cat("\n PA is the name of the parent node.")
  cat("\n N is the # of observations in the node.")
  cat("\n Yhat is the value representing the node.")
  cat("\n ImpMeasure is theimpurity measure of the node.\n")
  if( ncaty > 0 ) printm(Node, PA, N, ImpMeasure, Yhat, Prob )
  else  printm(Node, PA, N, ImpMeasure, Yhat )

  if( dump == 1 ){
   if( ncaty > 0 ) res=data.frame(Node, PA, N, ImpMeasure, Yhat, Prob)
   else res=data.frame(Node, PA, N, ImpMeasure, Yhat)

   return( res )
  }

} # print_leaf_dt


