#' Simplify the Conjunction of Inequalities
#'
#' @param ineq a string containing a conjunction of inequalities
#' @param print = 1 to print the result
#'
#' @details
#' The inequality of the form \code{aL < x aR} must precede the & sign as \cr
#' \code{simp_ineq("1 <= x < 5    &    x <= 3")}. \cr
#' It is assumed that the inequalities are valid.
#' Therefore, the order of the boundaries are not checked.
#'
#' @return A list of
#' \preformatted{
#' condc A string consisting of inequalities: the first term is the variable.
#' condl A vector consisting of aR, op1, x, op2, aL
#' }
#'
#'
#' @examples
#' res <- simp_ineq("x >= 1  &  x < 5")
#' res2 <- simp_ineq(paste(res$condc, "x < 3", sep=" & ") )
#' res3 <- simp_ineq(paste(res2$condc, "x > 2", sep=" & ") )
#' Print(res,res2,res3)
#'
#'
#' @export
#'

simp_ineq <- function( ineq, debug=0 ){
  # simplify inequalities
  # Shin-ichi Mayekawa
  # 20240216dnc,16,17
  # export: 20240218
  # a < x to x < a abandoned: 20240219
  # new ceq: 20240220dnc,21
  # invalid nn: 20240330
  #


  ############################# utility funcs ###########################
  ifagtb <- function( a, b ){
    as.numeric(a) > as.numeric(b)
  } # end of ifagtb
  ifaltb <- function( a, b ){
    as.numeric(a) < as.numeric(b)
  } # end of ifaltb


  ceq <- function( x, aL="", aR="", op1="", op2="" ){
   # returns
   #  aL op1 x op2 aR
   #  x op1 max(aL,aR)
   #  x > min(aL,aR)     # always put the variable first
   #  etc
   if( aL != ""  &  aR != "" ) txt=paste( aL, op1, x, op2, aR )
   else if( aL == ""  &  aR != "" ) txt=paste( x,op2,aR )
   else{
    # here aL != ""  &  aR == ""
    if( op1 %in% c("<","<=") ){
     if( op1 == "<" ) txt=paste( x, " > ", aL, sep="")
     else txt=paste( x, " >= ", aL, sep="")
    }
    else txt=paste( x, op1, aL, sep=" ")
   }
   if( debug >= 2 ) Print(txt)
   return( txt )
  } # end of ceq

  ############################# utility funcs ###########################


  #
  # This function assumes that the inequality is in the following form
  #   x < a   or   x <= a
  # where  x, to the left of inequality symbol, is the variable
  # and the a, to the right, is the limit.
  # Also, it can handle the inequality of the form
  #  aL < x < aR  &  x < a    where < can be <=
  # where the ones in between the inequality signs are assumed to be
  # the variable.
  #


  if( debug > 0 ) Print("\n\n***",ineq)

  if( regexpr("&",ineq) < 0 ){
    # no conjunctions

    if( debug > 0 ) Print("no conjunction",ineq)

#    Obsolete
#    # change a < x  to   x > a
#    op=regmatches(ineq,regexpr("(>|>=)",ineq))
#    if( length(op) > 0 ){
#      if( op == ">") op2="<"
#      else if( op == ">=" ) op2="<="
#      newexpr=paste("\\3 ",op2," \\1",sep="")
#      ineq=sub("([^<=]+)(>|>=)([^<=]+)",newexpr, ineq)
#    }
#

    x=op1=op2=aL=aR=""

    # decompose
    if( regexpr("[<>=]+", ineq) > 0 ){

      op=trim(regmatches(ineq,regexpr("[<>=]+",ineq)), all=1)
      temp=trim(unlist(strsplit(ineq,op,fixed=1)),leading=1)
      x=temp[1]
      if( op == ">"){
        if( debug > 0 ) cat("# x > a    -->  a < x . .\n")
        op1="<"
        aL=temp[2]; aR=""
      }
      else if( op == ">=" ){
        if( debug > 0 ) cat("# x >= a    -->  a <= x . .\n")
        op1="<="
        aL=temp[2]; aR=""
      }
      else{
        if( debug > 0 ) cat("# x < a    -->  . .  x < a\n")
       op2=op
       aR=temp[2]; aL=""
      }

    }

    if( debug > 0 ){
     Print( op,temp )
     Print(aL,op1,x,op2, aR)
     Print("condition =", ineq)
    }

    condl=c( aL,op1, x, op2, aR)
    names(condl)=c( "aL","op1", "x", "op2", "aR" )
    res=list( condc=ineq, condl=condl )
    return( res )

  }
  else{
    # conjunction found
    #
    # assuming that the variable name precedes the inequality sign.
    #

    EQ=unlist(strsplit(ineq," & "))
    eq1=EQ[1]
    eq2=EQ[2]

    #
    # check eq1 for two < symbols such as:
    # ineq = EQ = "aR <= x < aL   &   x > b"
    #       eq1 = "aR <= x < aL"
    #

    if( debug > 0 )  Print( "conjunction found", EQ,eq1,eq2 )


    if( regexpr("[^<=]+(<|<=)[^<=]+(<|<=)[^<=]+", eq1) > 0 ){

      # double < found, i.e. eq1 is of the form  aL < x < aR.

      op2=regmatches(eq2,regexpr("[<>=]+",eq2))
      temp2=trim(unlist(strsplit(eq2,op2,fixed=1)),leading=1)

      opop=unlist( regmatches(eq1,gregexpr("[<=]+",eq1)) )
      op1L=opop[1]; op1R=opop[2]

      temp1=trim(unlist(strsplit(eq1,"[<=]+",fixed=0)),leading=1)
      aL=temp1[1]
      x=temp1[2]
      aR=temp1[3]

      if( !(x == temp2[1]) ){
        return( ineq )
      }

      b=temp2[2]

     if( debug ){
      cat("double < found. i.e. eq1 is of the form  aL < x < aR")
      Print(aL,op1L,x,op1R,aR,",",x,op2, b)
     }

      if( op2 %in% c("<=","<") ){
        if( debug > 0 ) cat("# aL < x < aR  &  x < b \n")
        if( ifaltb(b,aR) ){
          if( debug > 0 ) cat("# b < aR,   aL < x < b \n")
          aL=aL; aR=b; op1=op1L; op2=op2
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("# b >= aR,  aL < x < aR \n")
          aL=aL; aR=b; op1=op1L; op2=op1R
          res=eq1
        }
      }
      else if( op2 %in% c(">=",">") ){
        if( debug > 0 ) cat("# aL < x < aR  &  x > b \n")
        if( op2 == ">=" ) op22="<="
        else op22="<"
        if( ifaltb(aL,b) ){
          if( debug > 0 ) cat("# aL < b,   b < x < aR \n")
          aL=b; aR=aR; op1=op22; op2=op1R
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("# aL >= b,   aL < x < aR \n")
          op1=op1L; op2=op1R
          res=ceq( x, aL, aR, op1, op2 )
        }
      }
      else{
        res="Invalid 1"
      }

      if( debug > 0 ){
        Print(aL,op1,x,op2, aR)
        Print("condition =", res)
      }

      condl=c( aL,op1, x, op2, aR)
      names(condl)=c( "aL","op1", "x", "op2", "aR" )
      res=list( condc=res, condl=condl )
      return( res )

    } # end of double <

    #
    # Here, ineq is of the form x > a  &  x > b
    #

    op1=trim(regmatches(eq1,regexpr("[<>=]+",eq1)), all=1)
    op2=trim(regmatches(eq2,regexpr("[<>=]+",eq2)), all=1)

    temp1=trim(unlist(strsplit(eq1,op1,fixed=1)),leading=1)
    temp2=trim(unlist(strsplit(eq2,op2,fixed=1)),leading=1)

    OPS=c("<=","<",">=",">")

    # locop1=which(op1==OPS)
    # locop2=which(op2==OPS)

    if( debug > 0 ) Print("Two simple inequalities found:",ineq)
    if( debug > 0 ) Print(eq1,eq2, op1,op2, temp1,temp2)

    x=temp1[1]
    if( !(x == temp2[1]) ){
      return( ineq )
    }

    a=temp1[2]
    b=temp2[2]

    if( op1 == "<=" ){
      if( op2 == "<=" ){
        if( debug > 0 ) cat("# x <= min(a,b) \n") ############################
        aR=min(as.numeric(a), as.numeric(b))
        aL=""; op2="<="; op1=""
        res=ceq( x, aL, aR, op1, op2 )
      } else if( op2 == "<" ){
        if( debug > 0 ) cat("# if( a < b ) x <= a; else x < b \n")
        if( ifaltb(a,b) ){
          if( debug > 0 ) cat("a < b \n")
          aL=a; aR=""; op1="<="; op2=""
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("a >= b \n")
          aL=a; op1="<"; aR=""; op2=""
          res=ceq( x, aL, aR, op1, op2 )
        }
      } else if( op2 == ">L="){
        if( debug > 0 ) cat("# if( a > b ) b <= x <= a; else error \n")
        if( ifagtb(a,b) ){
          if( debug > 0 ) cat("a > b \n")
          aL=b; aR=a; op1="<="; op2="<="
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("a <= b \n")
          aL=""; aR=""; op1=""; op2=""
          res="Invalid 2"
        }
      } else if( op2 == ">" ){
        if( debug > 0 ) cat("# if( a > b) b < x <= a; else error \n")
        if( ifagtb(a,b) ){
          if( debug > 0 ) cat("# a > b \n")
          aL=b; aR=a; op1="<"; op2="<="
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("# a <= b \n")
          aL=""; aR=""; op1=""; op2=""
          res="Invalid 3"
        }
      }
      else{
        aL=""; aR=""; op1=""; op2=""
        res="Invalid 4"
      }
    } else if( op1 == "<"){
      if( op2 == "<=" ){
        if( debug > 0 ) cat("# if( a < b ) x < a; else x <= b \n")
        if( ifaltb(a,b) ){
          if( debug > 0 ) cat("a < b \n")
          aL=a; op1="<"; aR=""; op2=""
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("a >= b \n")
          aL=b; op1="<="; aR=""; op2=""
          res=ceq( x, aL, aR, op1, op2 )
        }
      } else if( op2 == "<" ){
        if( debug > 0 ) cat("# x < min(a,b) \n")
        aR=min(as.numeric(a),as.numeric(b))
        aL="";  op2="<"; op1=""
        res=ceq( x, aL, aR, op1, op2 )
      } else if( op2 == ">="){
        if( debug > 0 ) cat("# if( a > b ) b <= x < a; else error \n")
        if( ifagtb(a,b) ){
          if( debug > 0 ) cat("a > b \n")
          aL=b; aR=a; op1="<="; op2="<"
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("a <= b \n")
          aL=""; aR=""; op1=""; op2=""
          res="Invalid 5"

        }
      } else if( op2 == ">" ){
        if( debug > 0 ) cat("# if( a > b) b < x < a; else error \n")
        if( ifagtb(a,b) ){
          if( debug > 0 ) cat("a > b \n")
          aL=b; aR=a; op1="<"; op2="<"
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("a <= b \n")
          aL=""; aR=""; op1=""; op2=""
          res="Invalid 6"

        }
      }
      else{
        aL=""; aR=""; op1=""; op2=""
        res="Invalid 7"
      }
    } else if( op1 == ">=" ){
      if( op2 == "<=" ){
        if( debug > 0 ) cat("# if( a < b ) a <= x <= b; else error \n")
        if( ifaltb(a,b) ){
          if( debug > 0 ) cat("a < b \n")
          aL=a; aR=b; op1="<="; op2="<="
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("a >= b \n")
          aL=""; aR=""; op1=""; op2=""
          res="Invalid 8"
        }
      } else if( op2 == "<" ){
        if( debug > 0 ) cat("# if( a < b ) a <= x < b; else error \n")
        if( ifaltb(a,b) ){
          if( debug > 0 ) cat("a < b \n")
          aL=a; aR=b; op1="<="; op2="<"
          res=ceq( x, aL, aR, op1, op2)
        }
        else{
          if( debug > 0 ) cat("a >= b \n")
          aL=""; aR=""; op1=""; op2=""
          res="Invalid 9"
        }
      } else if( op2 == ">="){
        if( debug > 0 ) cat("# x >= max(a,b) \n")
        aL=max(as.numeric(a), as.numeric(b))
        aR=""; op1=">="; op2=""
        res=ceq( x, aL, aR, op1, op2 )
        op1="<="
      } else if( op2 == ">" ){
        if( debug > 0 ) cat("# if( a > b ) x >= a; else x > b \n")
        if( ifagtb(a,b) ){
          if( debug > 0 ) cat("a > b \n")
          aL=a; op1=">="; op2=""
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("a <= b \n")
          aL=b; op1=">"; aR=""; op2=""
          res=ceq( x, aL, aR, op1, op2 )
        }
      }
      else{
        aL=""; aR=""; op1=""; op2=""
        res="Invalid 10"
      }
    } else if( op1 == ">" ){
      if( op2 == "<=" ){
        if( debug > 0 ) cat("# if( a < b ) a < x <= b; else error \n")
        if( ifaltb(a,b) ){
          if( debug > 0 ) cat("a < b \n")
          aL=a; aR=b; op1="<"; op2="<="
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("a >= b \n")
          aL=""; aR=""; op1=""; op2=""
          res="Invalid 11"
        }
      } else if( op2 == "<" ){
        if( debug > 0 ) cat("# if( a < b ) a < x < b; else error \n")
        if( ifaltb(a,b) ){
          if( debug > 0 ) cat("a < b \n")
          aL=a; aR=b; op1="<"; op2="<"
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("a >= b \n")
          aL=""; aR=""; op1=""; op2=""
          res="Invalid 12"
        }
      } else if( op2 == ">="){
        if( debug > 0 ) cat("# if( a > b ) x > a; else x >= b \n")
        if( ifagtb(a,b) ){
          if( debug > 0 ) cat("a > b \n")
          aL=a; op1=">"; aR=""; op2=""
          res=ceq( x, aL, aR, op1, op2 )
        }
        else{
          if( debug > 0 ) cat("a <= b \n")
          aL=b; op1=">="; aR=""; op2=""
          res=ceq( x, aL, aR, op1, op2 )
        }
      } else if( op2 == ">" ){
        if( debug > 0 ) cat("# x > max(a,b) \n")
        aL=max(as.numeric(a), as.numeric(b))
        aR=""; op1=">"; op2=""
        res=ceq( x, aL, aR, op1, op2 )
        op1="<"
      }
      else{
        aL=""; aR=""; op1=""; op2=""
        res="Invalid 13"
      }
    } else{
      aL=""; aR=""; op1=""; op2=""
      res="Invalid 14"
    }

  } # end of conjunction

  if( debug > 0 ){
    Print(aL,op1,x,op2, aR)
    Print("condition =", res)
  }

  condl=c( aL,op1, x, op2, aR )
  names(condl)=c( "aL","op1", "x", "op2", "aR" )
  res=list( condc=res, condl=condl )
  return( res )


} # end of simp_ineq



