#' Calculate Summary Statistics of a Decision Tree
#'
#' This function calculates various fit statistics given
#' the class DecTree object with or without a new regressor matrix. \cr
#' Japanese help file: \link[lazy.tree]{summary_dt_JPH}
#'
#' @param obj The output of DecTree or list of nodes.
#' @param y The criterion
#' @param X The NEW regressor matrix
#' @param yhat predicted value
#' @param prob probability of predicted value
#' @param minp Minimum value of probability
#' @param nleaf # of leaves when obj is not a class DecTree object
#' @param outpred = 1 to output the predicted values
#' @param print = 0 to suppress the output
#' @param title The title string
#'
#' @details
#' A set of summary statistics will be calculated using a pair of vectors 
#' \code{(y,yhat)} where \code{y} or \code{yhat} may come from \code{obj}
#' or given explicitly.
#' 
#' When \code{y} and \code{yhat} are present, it will be used to calculate the
#' summary statistics regardless of the presence of \code{obj}.
#' 
#' When \code{X} and \code{y} are given, the predicted values will be calculated
#' by \code{predict_dt} using the rules contained in \code{obj}.
#' 
#' When neither \code{X} nor \code{yhat} are given, \code{yhat} will be
#' picked up from the node information contained in \code{obj}
#' without using  \code{predict_dt}.
#' 

#'
#'
#' @return A list of the following:
#' \preformatted{
#' When \code{y} is categorical
#' nleaf:  # of leaves
#' Tab:       cross table: (row=true, col=predicted)
#' margin :   marginal frequencies
#' Gini: gini coefficient
#' mllh: minus log multinomial likelihood
#' nparam: # of parameters
#' aic: aic
#' bic: bic
#' tprecision: total precision
#' macroF1: macro F1
#' weightedF1: # of obs. weighted F1
#' SCT: a matrix of accuracy, precision, recall, specificity, and F1-score
#' for each levels of the criterion.
#'
#' When \code{y} is not categorical
#' nleaf:  # of leaves
#' corr:      correlation coefficient
#' }
#'
#' When \code{outpred=1}, \code{yhat, prob, Node} will be output.
#'
#'
#' @examples
#' Xc <- as.data.frame(cbind(square$c1,square$c2))
#' res1 <- DecTree( square$Class4, Xc, maxdepth=3, print=1 )
#' # temp and temp2 should be the same since Xc is not a new regressor matrix.
#' # No X is given.
#' temp <- summary_dt( res1, square$Class4, title="square" )
#' # The same X as contained in obj is given.
#' temp2 <- summary_dt( res1, square$Class4, Xc, title="square" )
#' # from y and yhat, no probs
#' temp22 <- summary_dt( yhat=res1$yhat, y=square$Class4)
#' # new X is given.
#' temp3 <- summary_dt( res1, square$Class4[1:10], Xc[1:10,], title="square" )
#'
#' @export
#'


summary_dt <- function( obj=NULL, y=NULL, X=NULL, yhat=NULL, prob=NULL
                        , minp=1e-4, nleaf=0, print=1
                        , outpred=0, title="" ){
 # calculate cross table from node and predicted values
 # Shin-ichi Mayekawa
 # 20230907
 # X,y -> y,X: 20230909
 # corr: 20230909
 # use subscript: 20230914
 # bugfix: 20230915,16
 # vertex --> node: 20230928
 # cross_table --> cross_table_dt: 20231002
 # DecTree object: 20231016
 # errRowm Col: 20231108
 # return the result: 20231108
 # cross_table_dt renamed as summary_dt: 20231109
 # weighted F1: 20231110cot
 # roxygen2: 20231121dnc
 # use minp when X is given: 20231122
 # outpred: 20231123
 # aic and nleaf: 20231123
 # Gini: 20231125
 # new entropy: 20231127
 # Gini divided by n: 20240227dnc
 # majestic < 0: 20240501
 # yhat: 20240507
 # get_valLL method=2: 20240518,23,24
 # y from obj: 20240518,23,24
 #

 
 
 calc_summary <- function( y, yhat ){
  # calculation of fit statistics
  # Shin-ichi Mayekawa
  # 20230907
  # core part converted to a function: 20240523
  # 

  
  # clear
  Tab=NULL; margin=NULL; corr=NULL
  
  
  #
  # continuous y
  #
  
  
  if( !is.factor(y) ){
   
   r=cor(y,yhat )
   
   if( print > 0 ){
    cat("\n", title, "\n", sep="")
    cat("\nCorrelation Coefficient between obs y and pred y =", r,"\n")
   }
   
   if( outpred == 1 ){
    res=named_list( Tab=Tab, margin, corr=r, y=y, yhat=yhat, prob=NA
                    , Node, nleaf )
   }
   else{
    res=named_list( Tab=Tab, margin, corr=r, nleaf )
   }
   
   return( res )
   
  } # end of continuous y
  
  
  #
  # categorical y
  #
  
  # # of categories (category levels) of y
  n=length(y)
  if( is.factor(y) ) ncaty=length(levels(y))
  else ncaty=0
  
  classnam=levels(y)
  tab=unclass( table(y, yhat) )
  Tab=matrix(0,ncaty,ncaty)
  colnames(Tab)=classnam; rownames(Tab)=classnam
  loc=classnam %in% colnames(tab)
  Tab[,loc]=tab
  
  TT=matrix(0,ncaty,4)
  colnames(TT)=c("TP","FP","TN","FN")
  SCT=matrix(0,ncaty,5)
  colnames(SCT)=c("accuracy","precision"
                  ,"recall","specificity","F1-score")
  for( i in 1:ncaty ){
   TP=Tab[i,i]
   FP=sum(Tab[-i,i])
   TN=sum(Tab[-i,-i])
   FN=sum(Tab[i,-i])
   SCT[i,1]=(TP+TN)/(TP+TN+FP+FN) # Accuracy
   SCT[i,2]=TP/(TP+FP)            # PPV (positive predictive value) or Precision
   SCT[i,3]=TP/(TP+FN) #TPR (true-positive rate) or Sensitivity, Recall, HitRate
   SCT[i,4]=TN/(TN+FP)            # TNR (true negative rate) or Specificity
   SCT[i,5]=2*TP/(2*TP+FP+FN)     # F1
  }
  
  errRow=rowSums(Tab-diag(diag(Tab)))/rowSums(Tab)
  errCol=colSums(Tab-diag(diag(Tab)))/colSums(Tab)
  rowSTab=rowSums(Tab)
  colSTab=colSums(Tab)
  margin=cbind(rowSTab, colSTab)
  colnames(margin)=c("rowsum","colsum")
  rownames(margin)=classnam
  
  
  tTP=sum(diag(Tab))
  tFP=sum(Tab-diag(diag(Tab)))
  tprecision=tTP/(tTP+tFP)
  macroF1=mean(SCT[,"F1-score"])
  weightedF1=weighted.mean(SCT[,"F1-score"],rowSTab)
  
  
  if( !is.null(prob) ){
   # Gini
   Gini=sum(prob*(1-prob))/n
   
   # mllh
   #
   # take care of small probs
   locz=which(prob <= 0)
   prob1=prob
   prob1[locz]=10
   logp=log(prob1)
   logp[locz]=0
   
   loc=cbind(1:n,as.numeric(y))
   # mllh=-sum(log(prob1[loc]))
   mllh=-sum(logp[loc])
   nparam=nleaf*(ncaty-1)
   aic=2*(mllh+nparam)
   bic=2*mllh+nparam*log(n)
  }
  else{
   Gini=NA; mllh=NA; nparam=NA; aic=NA; bic=NA
  }
  
  if( print > 0 ){
   cat("\n", title, "\n", sep="")
   cat("\nCross Table from Training Data:  row=true, col=predicted\n")
   Print(Tab, errRow, errCol)
   Print(margin)
   Print(SCT)
   cat("* recall = True Positive Rate,  specificity = True Negative Rate\n")
   Print(ncaty, nleaf)
   Print(mllh, nparam, aic, bic)
   Print(Gini, tprecision, macroF1, weightedF1, fmt=".4 5.4")
  }
  
  if( outpred == 1 ){
   res=named_list( Tab, margin, cor=NA, Gini, mllh, aic, bic, nparam, tprecision
                   , macroF1, weightedF1, SCT, y, yhat, prob, Node, nleaf )
  }
  else{
   res=named_list( Tab, margin, cor=NA, Gini, mllh, aic, bic, nparam
                   , tprecision, macroF1, weightedF1, SCT, nleaf )
  }
  
  return( res )
  
 } # end of calc_summary
 
 
 
 
 if( !is.null(y)  && !is.null(yhat) ){
  
  res=calc_summary( y, yhat )
  return( res )
  
 }
 
 
 
 #
 # Here, either y or yhat is not given
 # 
 
 # check if obj is given
 if( is.null(obj) ){
  cat("\nerror1:(summary_dt) obj must be given when (y,yhat) is missing.\n")
  stop()
 }
 
 #
 # Here, obj is given
 # 
 
 # # of categories (category levels) of y
 n=length(y)
 if( is.factor(y) ) ncaty=length(levels(y))
 else ncaty=0


 # check if yhat is given
 if( is.null(yhat) ){

    # DecTree object
    if( class(obj) == "DecTree" ){
      node=obj$node
      nleaf=obj$nleaf
      majestic=obj$majestic
      ncaty=obj$ncaty
      n=obj$n
      if( majestic > 0 ) rm(obj)
    }
    else if( is.list(obj) & !is.data.frame(obj) ){
      if( nleaf == 0 ){
        cat(
          "\nerror1:(summary_dt) nleaf must be given when class(obj) !="
          , "'Dectree'\n")
      }
      nam=names(obj[[1]])
      if( all(
        c("parent","set","depth","nobs","Left","Right","j") %in% names(obj[[1]])
      ) ){
        node=obj
        rm(obj)
      }
    }
    else{
      cat("\nerror1:(summary_dt) The input object is not a DecTree object.\n")
      stop()
    }
  


    Nobs=get_valLL(node,"nobs", method=2)

    # predicted value
    if( !is.null(X) ){
     
      if( is.null(y) ){
       cat("\nerror1:(summary_dt) When X is given, y must be given.\n")
       stop()
      }

      if( majestic > 0 ) temp=predict_dt( X, node )
      else{
        temp=predict_dt( X, obj )
        rm(obj)
      }
      yhat=temp$yhat
      Node~temp$Node
      prob=temp$prob
      n=nrow(X)

    }
    else{

     # For each leaf node, pick up yhat and prob
      
     if( is.null(y) ) y=obj$y
     
      n=length(y)
      yhat=numeric(n)
      Node=numeric(n)
      id=numeric(n)
      prob=matrix(0,n,ncaty)
      i=0
      for( h in 1:length(node) ){
        # Print(h,i,node[[h]]$subset,node[[h]]$yhat)
        if( node[[h]]$j == 0 ){
          subset=node[[h]]$set
          lens=length(subset)
          yhat[(i+1):(i+lens)]=rep(node[[h]]$yhat, lens)
          id[(i+1):(i+lens)]=subset
          Node[(i+1):(i+lens)]=h
          if( ncaty > 0 ){
            pp=node[[h]]$prob
            #     if( any(pp <= minp) ){
            #      pp[pp<=minp]=minp
            #      pp=pp/sum(pp)
            #     }
            prob[(i+1):(i+lens),]=matrix(pp, lens, ncaty, byrow=1)
          }
          i=i+lens
        }
      } # end of h

      # sort back
      od=order(id)
      yhat=yhat[od]
      Node=Node[od]
      prob[]=prob[od,]

    }

    # printm(y,yhat, prob)


  } # end of yhat not given
  else if( is.null(y) ){
   
   cat("\nerror1:(summary_dt) y must be given when yhat is given.\n")
   stop()
   
  }

 
 
 #
 # Here, (y, yhat) is available only which we need to calculate the summary.
 # 
 
 
 res=calc_summary( y, yhat )
 
 
 return( res )

} # end of summary_dt



